"""HEC-RAS Plan HDF class."""

from .geom import RasGeomHdf
from .utils import (
    df_datetimes_to_str,
    ras_timesteps_to_datetimes,
    parse_ras_datetime_ms,
    deprecated,
    convert_ras_hdf_value,
)

from geopandas import GeoDataFrame
import h5py
import numpy as np
from pandas import DataFrame
import pandas as pd
import xarray as xr

from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional, Tuple, Union


class RasPlanHdfError(Exception):
    """HEC-RAS Plan HDF error class."""

    pass


class XsSteadyOutputVar(Enum):
    """Summary of steady cross section output variables."""

    ENERGY_GRADE = "Energy Grade"
    FLOW = "Flow"
    WATER_SURFACE = "Water Surface"
    ENCROACHMENT_STATION_LEFT = "Encroachment Station Left"
    ENCROACHMENT_STATION_RIGHT = "Encroachment Station Right"
    AREA_INEFFECTIVE_TOTAL = "Area including Ineffective Total"
    VELOCITY_TOTAL = "Velocity Total"


XS_STEADY_OUTPUT_ADDITIONAL = [
    XsSteadyOutputVar.ENCROACHMENT_STATION_LEFT,
    XsSteadyOutputVar.ENCROACHMENT_STATION_RIGHT,
    XsSteadyOutputVar.AREA_INEFFECTIVE_TOTAL,
    XsSteadyOutputVar.VELOCITY_TOTAL,
]


class SummaryOutputVar(Enum):
    """Summary output variables."""

    MAXIMUM_WATER_SURFACE = "Maximum Water Surface"
    MINIMUM_WATER_SURFACE = "Minimum Water Surface"
    MAXIMUM_FACE_VELOCITY = "Maximum Face Velocity"
    MINIMUM_FACE_VELOCITY = "Minimum Face Velocity"
    CELL_MAXIMUM_WATER_SURFACE_ERROR = "Cell Maximum Water Surface Error"
    CELL_CUMULATIVE_ITERATION = "Cell Cumulative Iteration"
    CELL_LAST_ITERATION = "Cell Last Iteration"


SUMMARY_OUTPUT_VARS_CELLS = [
    SummaryOutputVar.MAXIMUM_WATER_SURFACE,
    SummaryOutputVar.MINIMUM_WATER_SURFACE,
    SummaryOutputVar.CELL_MAXIMUM_WATER_SURFACE_ERROR,
    SummaryOutputVar.CELL_CUMULATIVE_ITERATION,
    SummaryOutputVar.CELL_LAST_ITERATION,
]

SUMMARY_OUTPUT_VARS_FACES = [
    SummaryOutputVar.MAXIMUM_FACE_VELOCITY,
    SummaryOutputVar.MINIMUM_FACE_VELOCITY,
]


class TimeSeriesOutputVar(Enum):
    """Time series output variables."""

    # Default Outputs
    WATER_SURFACE = "Water Surface"
    FACE_VELOCITY = "Face Velocity"

    # Optional Outputs
    CELL_COURANT = "Cell Courant"
    CELL_CUMULATIVE_PRECIPITATION_DEPTH = "Cell Cumulative Precipitation Depth"
    CELL_DIVERGENCE_TERM = "Cell Divergence Term"
    CELL_EDDY_VISCOSITY_X = "Cell Eddy Viscosity - Eddy Viscosity X"
    CELL_EDDY_VISCOSITY_Y = "Cell Eddy Viscosity - Eddy Viscosity Y"
    CELL_FLOW_BALANCE = "Cell Flow Balance"
    CELL_HYDRAULIC_DEPTH = "Cell Hydraulic Depth"
    CELL_INVERT_DEPTH = "Cell Invert Depth"
    CELL_STORAGE_TERM = "Cell Storage Term"
    CELL_VELOCITY_X = "Cell Velocity - Velocity X"
    CELL_VELOCITY_Y = "Cell Velocity - Velocity Y"
    CELL_VOLUME = "Cell Volume"
    CELL_VOLUME_ERROR = "Cell Volume Error"
    CELL_WATER_SOURCE_TERM = "Cell Water Source Term"
    CELL_WATER_SURFACE_ERROR = "Cell Water Surface Error"

    FACE_COURANT = "Face Courant"
    FACE_CUMULATIVE_VOLUME = "Face Cumulative Volume"
    FACE_EDDY_VISCOSITY = "Face Eddy Viscosity"
    FACE_FLOW = "Face Flow"
    FACE_FLOW_PERIOD_AVERAGE = "Face Flow Period Average"
    FACE_FRICTION_TERM = "Face Friction Term"
    FACE_PRESSURE_GRADIENT_TERM = "Face Pressure Gradient Term"
    FACE_SHEAR_STRESS = "Face Shear Stress"
    FACE_TANGENTIAL_VELOCITY = "Face Tangential Velocity"
    FACE_WATER_SURFACE = "Face Water Surface"
    FACE_WIND_TERM = "Face Wind Term"


TIME_SERIES_OUTPUT_VARS_CELLS = [
    TimeSeriesOutputVar.WATER_SURFACE,
    TimeSeriesOutputVar.CELL_COURANT,
    TimeSeriesOutputVar.CELL_CUMULATIVE_PRECIPITATION_DEPTH,
    TimeSeriesOutputVar.CELL_DIVERGENCE_TERM,
    TimeSeriesOutputVar.CELL_EDDY_VISCOSITY_X,
    TimeSeriesOutputVar.CELL_EDDY_VISCOSITY_Y,
    TimeSeriesOutputVar.CELL_FLOW_BALANCE,
    TimeSeriesOutputVar.CELL_HYDRAULIC_DEPTH,
    TimeSeriesOutputVar.CELL_INVERT_DEPTH,
    TimeSeriesOutputVar.CELL_STORAGE_TERM,
    TimeSeriesOutputVar.CELL_VELOCITY_X,
    TimeSeriesOutputVar.CELL_VELOCITY_Y,
    TimeSeriesOutputVar.CELL_VOLUME,
    TimeSeriesOutputVar.CELL_VOLUME_ERROR,
    TimeSeriesOutputVar.CELL_WATER_SOURCE_TERM,
    TimeSeriesOutputVar.CELL_WATER_SURFACE_ERROR,
]

TIME_SERIES_OUTPUT_VARS_FACES = [
    TimeSeriesOutputVar.FACE_COURANT,
    TimeSeriesOutputVar.FACE_CUMULATIVE_VOLUME,
    TimeSeriesOutputVar.FACE_EDDY_VISCOSITY,
    TimeSeriesOutputVar.FACE_FLOW,
    TimeSeriesOutputVar.FACE_FLOW_PERIOD_AVERAGE,
    TimeSeriesOutputVar.FACE_FRICTION_TERM,
    TimeSeriesOutputVar.FACE_PRESSURE_GRADIENT_TERM,
    TimeSeriesOutputVar.FACE_SHEAR_STRESS,
    TimeSeriesOutputVar.FACE_TANGENTIAL_VELOCITY,
    TimeSeriesOutputVar.FACE_VELOCITY,
    TimeSeriesOutputVar.FACE_WATER_SURFACE,
    # TODO: investigate why "Face Wind Term" data gets written as a 1D array
    # TimeSeriesOutputVar.FACE_WIND_TERM,
]

TIME_SERIES_OUTPUT_VARS_DEFAULT = [
    TimeSeriesOutputVar.WATER_SURFACE,
    TimeSeriesOutputVar.FACE_VELOCITY,
]


class RasPlanHdf(RasGeomHdf):
    """HEC-RAS Plan HDF class."""

    PLAN_INFO_PATH = "Plan Data/Plan Information"
    PLAN_PARAMS_PATH = "Plan Data/Plan Parameters"
    PRECIP_PATH = "Event Conditions/Meteorology/Precipitation"
    OBS_DATA_PATH = "Event Conditions/Observed Data"
    RESULTS_UNSTEADY_PATH = "Results/Unsteady"
    RESULTS_UNSTEADY_SUMMARY_PATH = f"{RESULTS_UNSTEADY_PATH}/Summary"
    VOLUME_ACCOUNTING_PATH = f"{RESULTS_UNSTEADY_PATH}/Volume Accounting"
    BASE_OUTPUT_PATH = f"{RESULTS_UNSTEADY_PATH}/Output/Output Blocks/Base Output"
    SUMMARY_OUTPUT_2D_FLOW_AREAS_PATH = (
        f"{BASE_OUTPUT_PATH}/Summary Output/2D Flow Areas"
    )
    UNSTEADY_TIME_SERIES_PATH = f"{BASE_OUTPUT_PATH}/Unsteady Time Series"
    REFERENCE_LINES_OUTPUT_PATH = f"{UNSTEADY_TIME_SERIES_PATH}/Reference Lines"
    REFERENCE_POINTS_OUTPUT_PATH = f"{UNSTEADY_TIME_SERIES_PATH}/Reference Points"
    OBS_FLOW_OUTPUT_PATH = f"{OBS_DATA_PATH}/Flow"
    OBS_STAGE_OUTPUT_PATH = f"{OBS_DATA_PATH}/Stage"

    RESULTS_STEADY_PATH = "Results/Steady"
    BASE_STEADY_PATH = f"{RESULTS_STEADY_PATH}/Output/Output Blocks/Base Output"
    STEADY_PROFILES_PATH = f"{BASE_STEADY_PATH}/Steady Profiles"
    STEADY_XS_PATH = f"{STEADY_PROFILES_PATH}/Cross Sections"
    STEADY_XS_ADDITIONAL_PATH = f"{STEADY_XS_PATH}/Additional Variables"

    def __init__(self, name: str, **kwargs):
        """Open a HEC-RAS Plan HDF file.

        Parameters
        ----------
        name : str
            The path to the RAS Plan HDF file.
        kwargs : dict
            Additional keyword arguments to pass to h5py.File
        """
        super().__init__(name, **kwargs)

    def _simulation_start_time(self) -> datetime:
        """Return the simulation start time from the plan file.

        Returns
        -------
        datetime
            The simulation start time.
        """
        plan_info = self.get_plan_info_attrs()
        return plan_info["Simulation Start Time"]

    def _2d_flow_area_names_and_counts(self) -> List[Tuple[str, int]]:
        """
        Return a list of 2D flow area names and cell counts.

        Returns
        -------
        List[Tuple[str, int]]
            A list of tuples, where each tuple contains a 2D flow area name
            and the number of cells in that area.
        """
        d2_flow_areas = self[f"{self.FLOW_AREA_2D_PATH}/Attributes"][:]
        return [
            (d2_flow_area[0].decode("utf-8"), d2_flow_area[-1])
            for d2_flow_area in d2_flow_areas
        ]

    def _mesh_summary_output_group(
        self, mesh_name: str, output_var: SummaryOutputVar
    ) -> h5py.Group:
        """Return the HDF group for a 2D flow area summary output variable.

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.
        output_var : str
            The name of the output variable.

        Returns
        -------
        h5py.Group
            The HDF group for the output variable.
        """
        output_path = (
            f"{self.SUMMARY_OUTPUT_2D_FLOW_AREAS_PATH}/{mesh_name}/{output_var.value}"
        )
        output_group = self.get(output_path)
        if output_group is None:
            raise RasPlanHdfError(
                f"Could not find HDF group at path '{output_path}'."
                " Does the Plan HDF file contain 2D output data?"
            )
        return output_group

    def _mesh_summary_output_min_max_values(
        self, mesh_name: str, var: SummaryOutputVar
    ) -> np.ndarray:
        """Return values for a "Maximum"/"Minimum" summary output variable.

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.
        var : SummaryOutputVar
            The summary output variable to retrieve.

        Returns
        -------
        np.ndarray
            An array of maximum water surface elevation values.
        """
        max_ws_group = self._mesh_summary_output_group(mesh_name, var)
        max_ws_raw = max_ws_group[:]
        max_ws_values = max_ws_raw[0]
        return max_ws_values

    def _summary_output_min_max_time_unit(self, dataset: h5py.Dataset) -> str:
        """Return the time unit for "Maximum"/"Minimum" summary output datasets.

        I.e., for summary output such as "Maximum Water Surface", "Minimum Water Surface", etc.

        Should normally return the string: "days".

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.

        Returns
        -------
        str
            The time unit for the maximum water surface elevation data.
        """
        if "Units per row" in dataset.attrs:
            units = dataset.attrs["Units per row"]
        else:
            units = dataset.attrs["Units"]
        # expect an array of size 2, with the first element being length or velocity units
        # and the second element being time units (e.g., ["ft", "days"])
        time_unit = units[1]
        return time_unit.decode("utf-8")

    def _mesh_summary_output_min_max_times(
        self,
        mesh_name: str,
        var: SummaryOutputVar,
        time_unit: str = "days",
        round_to: str = "0.1 s",
    ) -> np.ndarray[np.datetime64]:
        """Return an array of times for min/max summary output data.

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.
        var : SummaryOutputVar
            The summary output variable to retrieve.
        time_unit : str, optional
            The time unit for the maximum water surface elevation data.
            Default: "days".
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).

        Returns
        -------
        np.ndarray[np.datetime64]
            An array of times for the maximum water surface elevation data.
        """
        start_time = self._simulation_start_time()
        max_ws_group = self._mesh_summary_output_group(mesh_name, var)
        time_unit = self._summary_output_min_max_time_unit(max_ws_group)
        max_ws_raw = max_ws_group[:]
        max_ws_times_raw = max_ws_raw[1]
        # we get weirdly specific datetime values if we don't round to e.g., 0.1 seconds;
        # otherwise datetimes don't align with the actual timestep values in the plan file
        max_ws_times = ras_timesteps_to_datetimes(
            max_ws_times_raw, start_time, time_unit=time_unit, round_to=round_to
        )
        return max_ws_times

    def _mesh_summary_output_min_max(
        self,
        var: SummaryOutputVar,
        value_col: str = "value",
        time_col: str = "time",
        round_to: str = "0.1 s",
    ) -> DataFrame:
        """Return the min/max values and times for a summary output variable.

        Valid for:
        - Maximum Water Surface
        - Minimum Water Surface
        - Maximum Face Velocity
        - Minimum Face Velocity
        - Cell Maximum Water Surface Error

        Parameters
        ----------
        var : SummaryOutputVar
            The summary output variable to retrieve.
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id' or 'face_id', 'value', and 'time'.
        """
        dfs = []
        for mesh_name, cell_count in self._2d_flow_area_names_and_counts():
            values = self._mesh_summary_output_min_max_values(mesh_name, var)
            times = self._mesh_summary_output_min_max_times(
                mesh_name, var, round_to=round_to
            )
            if var in [
                SummaryOutputVar.MAXIMUM_FACE_VELOCITY,
                SummaryOutputVar.MINIMUM_FACE_VELOCITY,
            ]:
                geom_id_col = "face_id"
            else:
                geom_id_col = "cell_id"
                # The 2D mesh output data contains values for more cells than are actually
                # in the mesh. The the true number of cells for a mesh is found in the table:
                # "/Geometry/2D Flow Areas/Attributes". The number of cells in the 2D output
                # data instead matches the number of cells in the "Cells Center Coordinate"
                # array, which contains extra points along the perimeter of the mesh. These
                # extra points are appended to the end of the mesh data and contain bogus
                # output values (e.g., 0.0, NaN). We need to filter out these bogus values.
                values = values[:cell_count]
                times = times[:cell_count]
            df = DataFrame(
                {
                    "mesh_name": [mesh_name] * len(values),
                    geom_id_col: range(len(values)),
                    value_col: values,
                    time_col: times,
                }
            )
            dfs.append(df)
        df = pd.concat(dfs, ignore_index=True)
        return df

    def _mesh_summary_output_basic(
        self, var: SummaryOutputVar, value_col: str = "value"
    ) -> DataFrame:
        """Return values and times for a summary output variable.

        Valid for:
        - Cell Cumulative Iteration (i.e. Cumulative Max Iterations)
        - Cell Last Iteration

        Parameters
        ----------
        var : SummaryOutputVar
            The summary output variable to retrieve.

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id' or 'face_id', 'value', and 'time'.
        """
        dfs = []
        for mesh_name, cell_count in self._2d_flow_area_names_and_counts():
            group = self._mesh_summary_output_group(mesh_name, var)
            values = group[:][:cell_count]
            df = DataFrame(
                {
                    "mesh_name": [mesh_name] * len(values),
                    "cell_id": range(len(values)),
                    value_col: values,
                }
            )
            dfs.append(df)
        df = pd.concat(dfs, ignore_index=True)
        return df

    def mesh_max_iter(self) -> DataFrame:
        """Return the number of times each cell in the mesh reached the max number of iterations.

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id', and 'max_iterations'.
        """
        df = self._mesh_summary_output_basic(
            SummaryOutputVar.CELL_CUMULATIVE_ITERATION, value_col="max_iter"
        )
        return df

    def mesh_last_iter(self) -> DataFrame:
        """Return the number of times each cell in the mesh was the last cell to converge.

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id', and 'last_iter'.
        """
        df = self._mesh_summary_output_basic(
            SummaryOutputVar.CELL_LAST_ITERATION, value_col="last_iter"
        )
        return df

    def mesh_max_ws(self, round_to: str = "0.1 s") -> DataFrame:
        """Return the max water surface elevation for each cell in the mesh.

        Includes the corresponding time of max water surface elevation.

        Parameters
        ----------
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
            See Pandas documentation for valid time units:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id', 'max_ws', and 'max_ws_time'.
        """
        df = self._mesh_summary_output_min_max(
            SummaryOutputVar.MAXIMUM_WATER_SURFACE,
            value_col="max_ws",
            time_col="max_ws_time",
            round_to=round_to,
        )
        return df

    def mesh_min_ws(self, round_to: str = "0.1 s") -> DataFrame:
        """Return the min water surface elevation for each cell in the mesh.

        Includes the corresponding time of min water surface elevation.

        Parameters
        ----------
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
            See Pandas documentation for valid time units:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id', 'min_ws', and 'min_ws_time'.
        """
        df = self._mesh_summary_output_min_max(
            SummaryOutputVar.MINIMUM_WATER_SURFACE,
            value_col="min_ws",
            time_col="min_ws_time",
            round_to=round_to,
        )
        return df

    def mesh_max_face_v(self, round_to: str = "0.1 s") -> DataFrame:
        """Return the max face velocity for each face in the mesh.

        Includes the corresponding time of max face velocity.

        Parameters
        ----------
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
            See Pandas documentation for valid time units:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'face_id', 'max_v', and 'max_v_time'.
        """
        df = self._mesh_summary_output_min_max(
            SummaryOutputVar.MAXIMUM_FACE_VELOCITY,
            value_col="max_v",
            time_col="max_v_time",
            round_to=round_to,
        )
        return df

    def mesh_min_face_v(self, round_to: str = "0.1 s") -> DataFrame:
        """Return the min face velocity for each face in the mesh.

        Includes the corresponding time of min face velocity.

        Parameters
        ----------
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
            See Pandas documentation for valid time units:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'face_id', 'min_v', and 'min_v_time'.
        """
        df = self._mesh_summary_output_min_max(
            SummaryOutputVar.MINIMUM_FACE_VELOCITY,
            value_col="min_v",
            time_col="min_v_time",
            round_to=round_to,
        )
        return df

    def mesh_max_ws_err(self, round_to: str = "0.1 s") -> DataFrame:
        """Return the max water surface error for each cell in the mesh.

        Includes the corresponding time of max water surface error.

        Parameters
        ----------
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
            See Pandas documentation for valid time units:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id', 'max_ws_err', and 'max_ws_err_time'.
        """
        df = self._mesh_summary_output_min_max(
            SummaryOutputVar.CELL_MAXIMUM_WATER_SURFACE_ERROR,
            value_col="max_ws_err",
            time_col="max_ws_err_time",
            round_to=round_to,
        )
        return df

    def mesh_summary_output(
        self, var: SummaryOutputVar, round_to: str = "0.1 s"
    ) -> DataFrame:
        """Return the summary output data for a given variable.

        Parameters
        ----------
        var : SummaryOutputVar
            The summary output variable to retrieve.

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id' or 'face_id', a value column,
            and a time column if the value corresponds to a specific time.
        """
        methods_with_times = {
            SummaryOutputVar.MAXIMUM_WATER_SURFACE: self.mesh_max_ws,
            SummaryOutputVar.MINIMUM_WATER_SURFACE: self.mesh_min_ws,
            SummaryOutputVar.MAXIMUM_FACE_VELOCITY: self.mesh_max_face_v,
            SummaryOutputVar.MINIMUM_FACE_VELOCITY: self.mesh_min_face_v,
            SummaryOutputVar.CELL_MAXIMUM_WATER_SURFACE_ERROR: self.mesh_max_ws_err,
        }
        other_methods = {
            SummaryOutputVar.CELL_CUMULATIVE_ITERATION: self.mesh_max_iter,
            SummaryOutputVar.CELL_LAST_ITERATION: self.mesh_last_iter,
        }
        if var in methods_with_times:
            df = methods_with_times[var](round_to=round_to)
        else:
            df = other_methods[var]()
        return df

    def _mesh_summary_outputs_df(
        self,
        cells_or_faces: str,
        output_vars: Optional[List[SummaryOutputVar]] = None,
        round_to: str = "0.1 s",
    ) -> DataFrame:
        if cells_or_faces == "cells":
            feature_id_field = "cell_id"
        elif cells_or_faces == "faces":
            feature_id_field = "face_id"
        else:
            raise ValueError('cells_or_faces must be either "cells" or "faces".')
        if output_vars is None:
            summary_output_vars = self._summary_output_vars(
                cells_or_faces=cells_or_faces
            )
        elif isinstance(output_vars, list):
            summary_output_vars = []
            for var in output_vars:
                if not isinstance(var, SummaryOutputVar):
                    var = SummaryOutputVar(var)
                summary_output_vars.append(var)
        else:
            raise ValueError(
                "include_output must be a boolean or a list of SummaryOutputVar values."
            )
        df = self.mesh_summary_output(summary_output_vars[0], round_to=round_to)
        for var in summary_output_vars[1:]:
            df_var = self.mesh_summary_output(var, round_to=round_to)
            df = df.merge(df_var, on=["mesh_name", feature_id_field], how="left")
        return df

    def mesh_cells_summary_output(self, round_to: str = "0.1 s") -> DataFrame:
        """
        Return a DataFrame with summary output data for each mesh cell in the model.

        Parameters
        ----------
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
            See Pandas documentation for valid time units:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'cell_id', and columns for each
            summary output variable.
        """
        return self._mesh_summary_outputs_df("cells", round_to=round_to)

    def mesh_faces_summary_output(self, round_to: str = "0.1 s") -> DataFrame:
        """
        Return a DataFrame with summary output data for each mesh face in the model.

        Parameters
        ----------
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
            See Pandas documentation for valid time units:
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html

        Returns
        -------
        DataFrame
            A DataFrame with columns 'mesh_name', 'face_id', and columns for each
            summary output variable.
        """
        return self._mesh_summary_outputs_df("faces", round_to=round_to)

    def _summary_output_vars(
        self, cells_or_faces: Optional[str] = None
    ) -> List[SummaryOutputVar]:
        """Return a list of available summary output variables from the Plan HDF file.

        Returns
        -------
        List[SummaryOutputVar]
            A list of summary output variables.
        """
        mesh_names_counts = self._2d_flow_area_names_and_counts()
        mesh_names = [mesh_name for mesh_name, _ in mesh_names_counts]
        vars = set()
        for mesh_name in mesh_names:
            path = f"{self.SUMMARY_OUTPUT_2D_FLOW_AREAS_PATH}/{mesh_name}"
            datasets = self[path].keys()
            for dataset in datasets:
                try:
                    var = SummaryOutputVar(dataset)
                except ValueError:
                    continue
                vars.add(var)
        if cells_or_faces == "cells":
            vars = vars.intersection(SUMMARY_OUTPUT_VARS_CELLS)
        elif cells_or_faces == "faces":
            vars = vars.intersection(SUMMARY_OUTPUT_VARS_FACES)
        return sorted(list(vars), key=lambda x: x.value)

    def _mesh_summary_outputs_gdf(
        self,
        geom_func: str,
        cells_or_faces: str = "cells",
        include_output: Union[bool, List[SummaryOutputVar]] = True,
        round_to: str = "0.1 s",
        datetime_to_str: bool = False,
    ) -> GeoDataFrame:
        """Return a GeoDataFrame with mesh geometry and summary output data.

        Parameters
        ----------
        geom_func : str
            The method name to call to get the mesh geometry.
        cells_or_faces : str, optional
            The type of geometry to include in the GeoDataFrame.
            Must be either "cells" or "faces". (default: "cells")
        include_output : Union[bool, List[SummaryOutputVar]], optional
            If True, include all available summary output data in the GeoDataFrame.
            If a list of SummaryOutputVar values, include only the specified summary output data.
            If False, do not include any summary output data.
            (default: True)
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
        datetime_to_str : bool, optional
            If True, convert datetime columns to strings. (default: False)
        """
        gdf = getattr(super(), geom_func)()
        if include_output is False:
            return gdf
        if include_output is True:
            summary_output_vars = self._summary_output_vars(
                cells_or_faces=cells_or_faces
            )
        elif isinstance(include_output, list):
            summary_output_vars = []
            for var in include_output:
                if not isinstance(var, SummaryOutputVar):
                    var = SummaryOutputVar(var)
                summary_output_vars.append(var)
        else:
            raise ValueError(
                "include_output must be a boolean or a list of SummaryOutputVar values."
            )
        if cells_or_faces == "cells":
            geom_id_col = "cell_id"
        elif cells_or_faces == "faces":
            geom_id_col = "face_id"
        else:
            raise ValueError('cells_or_faces must be either "cells" or "faces".')
        for var in summary_output_vars:
            df = self.mesh_summary_output(var, round_to=round_to)
            gdf = gdf.merge(df, on=["mesh_name", geom_id_col], how="left")
        if datetime_to_str:
            gdf = df_datetimes_to_str(gdf)
        return gdf

    def mesh_cell_points(
        self,
        include_output: Union[bool, List[SummaryOutputVar], List[str]] = True,
        round_to: str = "0.1 s",
        datetime_to_str: bool = False,
    ) -> GeoDataFrame:
        """Return the cell points for each cell in the mesh, including summary output.

        Parameters
        ----------
        include_output : Union[bool, List[SummaryOutputVar], List[str]], optional
            If True, include all available summary output data in the GeoDataFrame.
            If a list of SummaryOutputVar values, include only the specified summary output data.
            If a list of strings, include only the specified summary output data by name.
            If False, do not include any summary output data.
            (default: True)
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
        datetime_to_str : bool, optional
            If True, convert datetime columns to strings. (default: False)

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame with columns 'mesh_name', 'cell_id', 'geometry', and columns for each
            summary output variable.
        """
        return self._mesh_summary_outputs_gdf(
            "mesh_cell_points",
            "cells",
            include_output=include_output,
            round_to=round_to,
            datetime_to_str=datetime_to_str,
        )

    def mesh_cell_polygons(
        self,
        include_output: Union[bool, List[SummaryOutputVar], List[str]] = True,
        round_to: str = "0.1 s",
        datetime_to_str: bool = False,
    ) -> GeoDataFrame:
        """Return the cell polygons for each cell in the mesh, including output.

        Parameters
        ----------
        include_output : Union[bool, List[SummaryOutputVar], List[str]], optional
            If True, include all available summary output data in the GeoDataFrame.
            If a list of SummaryOutputVar values, include only the specified summary output data.
            If a list of strings, include only the specified summary output data by name.
            If False, do not include any summary output data.
            (default: True)
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
        datetime_to_str : bool, optional
            If True, convert datetime columns to strings. (default: False)

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame with columns 'mesh_name', 'cell_id', 'geometry', and columns for each
            summary output variable.
        """
        return self._mesh_summary_outputs_gdf(
            "mesh_cell_polygons",
            "cells",
            include_output=include_output,
            round_to=round_to,
            datetime_to_str=datetime_to_str,
        )

    def mesh_cell_faces(
        self,
        include_output: Union[bool, List[SummaryOutputVar], List[str]] = True,
        round_to: str = "0.1 s",
        datetime_to_str: bool = False,
    ) -> GeoDataFrame:
        """Return the cell faces for each cell in the mesh, including output.

        Parameters
        ----------
        include_output : Union[bool, List[SummaryOutputVar], List[str]], optional
            If True, include all available summary output data in the GeoDataFrame.
            If a list of SummaryOutputVar values, include only the specified summary output data.
            If a list of strings, include only the specified summary output data by name.
            If False, do not include any summary output data.
            (default: True)
        round_to : str, optional
            The time unit to round the datetimes to. Default: "0.1 s" (seconds).
        datetime_to_str : bool, optional
            If True, convert datetime columns to strings. (default: False)

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame with columns 'mesh_name', 'cell_id', 'geometry', and columns for each
            summary output variable.
        """
        return self._mesh_summary_outputs_gdf(
            "mesh_cell_faces",
            "faces",
            include_output=include_output,
            round_to=round_to,
            datetime_to_str=datetime_to_str,
        )

    def unsteady_datetimes(self) -> List[datetime]:
        """Return the unsteady timeseries datetimes from the plan file.

        Returns
        -------
        List[datetime]
            A list of datetimes for the unsteady timeseries data.
        """
        group_path = f"{self.UNSTEADY_TIME_SERIES_PATH}/Time Date Stamp (ms)"
        raw_datetimes = self[group_path][:]
        dt = [parse_ras_datetime_ms(x.decode("utf-8")) for x in raw_datetimes]
        return dt

    def _mesh_timeseries_output_values_units(
        self,
        mesh_name: str,
        var: TimeSeriesOutputVar,
    ) -> Tuple[np.ndarray, str]:
        path = self._mesh_timeseries_output_path(mesh_name, var.value)
        group = self.get(path)
        try:
            import dask.array as da

            # TODO: user-specified chunks?
            values = da.from_array(group, chunks=group.chunks)
        except ImportError:
            values = group[:]
        units = group.attrs.get("Units")
        if units is not None:
            units = units.decode("utf-8")
        return values, units

    def mesh_timeseries_output(
        self,
        mesh_name: str,
        var: Union[str, TimeSeriesOutputVar],
        truncate: bool = True,
    ) -> xr.DataArray:
        """Return the time series output data for a given variable.

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.
        var : TimeSeriesOutputVar
            The time series output variable to retrieve.
        truncate : bool, optional
            If True, truncate the number of cells to the listed cell count.

        Returns
        -------
        xr.DataArray
            An xarray DataArray with dimensions 'time' and 'cell_id'.
        """
        times = self.unsteady_datetimes()
        mesh_names_counts = {
            name: count for name, count in self._2d_flow_area_names_and_counts()
        }
        if mesh_name not in mesh_names_counts:
            raise ValueError(f"Mesh '{mesh_name}' not found in the Plan HDF file.")
        if isinstance(var, str):
            var = TimeSeriesOutputVar(var)
        values, units = self._mesh_timeseries_output_values_units(mesh_name, var)
        if var in TIME_SERIES_OUTPUT_VARS_CELLS:
            cell_count = mesh_names_counts[mesh_name]
            if truncate:
                values = values[:, :cell_count]
            else:
                values = values[:, :]
            id_coord = "cell_id"
        elif var in TIME_SERIES_OUTPUT_VARS_FACES:
            id_coord = "face_id"
        else:
            raise ValueError(f"Invalid time series output variable: {var.value}")
        da = xr.DataArray(
            values,
            name=var.value,
            dims=["time", id_coord],
            coords={
                "time": times,
                id_coord: range(values.shape[1]),
            },
            attrs={
                "mesh_name": mesh_name,
                "variable": var.value,
                "units": units,
                "hdf_path": self._mesh_timeseries_output_path(mesh_name, var.value),
            },
        )
        return da

    def _mesh_timeseries_output_path(self, mesh_name: str, var_name: str) -> str:
        return f"{self.UNSTEADY_TIME_SERIES_PATH}/2D Flow Areas/{mesh_name}/{var_name}"

    def _mesh_timeseries_outputs(
        self, mesh_name: str, vars: List[TimeSeriesOutputVar], truncate: bool = True
    ) -> xr.Dataset:
        datasets = {}
        for var in vars:
            var_path = f"{self.UNSTEADY_TIME_SERIES_PATH}/2D Flow Areas/{mesh_name}/{var.value}"
            if self.get(var_path) is None:
                continue
            da = self.mesh_timeseries_output(mesh_name, var, truncate=truncate)
            datasets[var.value] = da
        ds = xr.Dataset(datasets, attrs={"mesh_name": mesh_name})
        return ds

    def mesh_cells_timeseries_output(self, mesh_name: str) -> xr.Dataset:
        """Return the time series output data for cells in a 2D flow area mesh.

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.

        Returns
        -------
        xr.Dataset
            An xarray Dataset with DataArrays for each time series output variable.
        """
        ds = self._mesh_timeseries_outputs(mesh_name, TIME_SERIES_OUTPUT_VARS_CELLS)
        return ds

    @deprecated
    def mesh_timeseries_output_cells(self, mesh_name: str) -> xr.Dataset:
        """Return the time series output data for cells in a 2D flow area mesh.

        Deprecated: use mesh_cells_timeseries_output instead.

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.

        Returns
        -------
        xr.Dataset
            An xarray Dataset with DataArrays for each time series output variable.
        """
        return self.mesh_cells_timeseries_output(mesh_name)

    def mesh_faces_timeseries_output(self, mesh_name: str) -> xr.Dataset:
        """Return the time series output data for faces in a 2D flow area mesh.

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.

        Returns
        -------
        xr.Dataset
            An xarray Dataset with DataArrays for each time series output variable.
        """
        ds = self._mesh_timeseries_outputs(mesh_name, TIME_SERIES_OUTPUT_VARS_FACES)
        return ds

    @deprecated
    def mesh_timeseries_output_faces(self, mesh_name: str) -> xr.Dataset:
        """Return the time series output data for faces in a 2D flow area mesh.

        Deprecated: use mesh_faces_timeseries_output instead.

        Parameters
        ----------
        mesh_name : str
            The name of the 2D flow area mesh.

        Returns
        -------
        xr.Dataset
            An xarray Dataset with DataArrays for each time series output variable.
        """
        return self.mesh_faces_timeseries_output(mesh_name)

    def reference_timeseries_output(self, reftype: str = "lines") -> xr.Dataset:
        """Return timeseries output data for reference lines or points from a HEC-RAS HDF plan file.

        Parameters
        ----------
        reftype : str, optional
            The type of reference data to retrieve. Must be either "lines" or "points".
            (default: "lines")

        Returns
        -------
        xr.Dataset
            An xarray Dataset with reference line timeseries data.
        """
        if reftype == "lines":
            output_path = self.REFERENCE_LINES_OUTPUT_PATH
            abbrev = "refln"
        elif reftype == "points":
            output_path = self.REFERENCE_POINTS_OUTPUT_PATH
            abbrev = "refpt"
        else:
            raise ValueError('reftype must be either "lines" or "points".')
        reference_group = self.get(output_path)
        if reference_group is None:
            raise RasPlanHdfError(
                f"Could not find HDF group at path '{output_path}'."
                f" Does the Plan HDF file contain reference {reftype[:-1]} output data?"
            )
        reference_names = reference_group["Name"][:]
        names = []
        mesh_areas = []
        for s in reference_names:
            name, mesh_area = s.decode("utf-8").split("|")
            names.append(name)
            mesh_areas.append(mesh_area)

        times = self.unsteady_datetimes()

        das = {}
        for var in ["Flow", "Velocity", "Water Surface"]:
            group = reference_group.get(var)
            if group is None:
                continue
            try:
                import dask.array as da

                # TODO: user-specified chunks?
                values = da.from_array(group, chunks=group.chunks)
            except ImportError:
                values = group[:]
            units = group.attrs["Units"].decode("utf-8")
            da = xr.DataArray(
                values,
                name=var,
                dims=["time", f"{abbrev}_id"],
                coords={
                    "time": times,
                    f"{abbrev}_id": range(values.shape[1]),
                    f"{abbrev}_name": (f"{abbrev}_id", names),
                    "mesh_name": (f"{abbrev}_id", mesh_areas),
                },
                attrs={"units": units, "hdf_path": f"{output_path}/{var}"},
            )
            das[var] = da
        return xr.Dataset(das)

    def reference_lines_timeseries_output(self) -> xr.Dataset:
        """Return timeseries output data for reference lines from a HEC-RAS HDF plan file.

        Returns
        -------
        xr.Dataset
            An xarray Dataset with timeseries output data for reference lines.
        """
        return self.reference_timeseries_output(reftype="lines")

    def observed_timeseries_input(self, vartype: str = "Flow") -> dict:
        """Return observed timeseries input data for reference lines and points from a HEC-RAS HDF plan file.

        Parameters
        ----------
        vartype : str, optional
            The type of observed data to retrieve. Must be either "Flow" or "Stage".
            (default: "Flow")

        Returns
        -------
        xr.Dataset
            An xarray Dataset with observed timeseries input data for both reference lines and reference points.
        """
        if vartype == "Flow":
            output_path = self.OBS_FLOW_OUTPUT_PATH
        elif vartype == "Stage":
            output_path = self.OBS_STAGE_OUTPUT_PATH
        else:
            raise ValueError('vartype must be either "Flow" or "Stage".')

        observed_group = self.get(output_path)
        if observed_group is None:
            raise RasPlanHdfError(
                f"Could not find HDF group at path '{output_path}'."
                f" Does the Plan HDF file contain reference {vartype} output data?"
            )
        if "Attributes" in observed_group.keys():
            attr_path = observed_group["Attributes"]
            attrs_df = pd.DataFrame(attr_path[:]).map(convert_ras_hdf_value)

        das = {}
        for idx, site in enumerate(observed_group.keys()):
            if site != "Attributes":
                # Site Ex: 'Ref Point: Grapevine_Lake_RP'
                site_path = observed_group[site]
                site_name = site.split(":")[1][1:]  # Grapevine_Lake_RP
                ref_type = site.split(":")[0]  # Ref Point
                if ref_type == "Ref Line":
                    ref_type = "refln"
                else:
                    ref_type = "refpt"
                df = pd.DataFrame(site_path[:]).map(convert_ras_hdf_value)
                # rename Date to time
                df = df.rename(columns={"Date": "time"})
                # Ensure the Date index is unique
                df = df.drop_duplicates(subset="time")
                # Package into an 1D xarray DataArray
                values = df["Value"].values
                times = df["time"].values
                da = xr.DataArray(
                    values,
                    name=vartype,
                    dims=["time"],
                    coords={
                        "time": times,
                    },
                    attrs={
                        "hdf_path": f"{output_path}/{site}",
                    },
                )
                # Expand dimensions to add additional coordinates
                da = da.expand_dims({f"{ref_type}_id": [idx - 1]})
                da = da.expand_dims({f"{ref_type}_name": [site_name]})
                das[site_name] = da
        das = xr.concat([das[site] for site in das.keys()], dim="time")
        return das

    def reference_points_timeseries_output(self) -> xr.Dataset:
        """Return timeseries output data for reference points from a HEC-RAS HDF plan file.

        Returns
        -------
        xr.Dataset
            An xarray Dataset with timeseries output data for reference points.
        """
        return self.reference_timeseries_output(reftype="points")

    def reference_summary_output(self, reftype: str = "lines") -> DataFrame:
        """Return summary output data for reference lines or points from a HEC-RAS HDF plan file.

        Returns
        -------
        DataFrame
            A DataFrame with reference line summary output data.
        """
        if reftype == "lines":
            abbrev = "refln"
        elif reftype == "points":
            abbrev = "refpt"
        else:
            raise ValueError('reftype must be either "lines" or "points".')
        ds = self.reference_timeseries_output(reftype=reftype)
        result = {
            f"{abbrev}_id": ds[f"{abbrev}_id"],
            f"{abbrev}_name": ds[f"{abbrev}_name"],
            "mesh_name": ds.mesh_name,
        }
        vars = {
            "Flow": "q",
            "Water Surface": "ws",
            "Velocity": "v",
        }
        for var, abbrev in vars.items():
            if var not in ds:
                continue
            max_var = ds[var].max(dim="time")
            max_time = ds[var].time[ds[var].argmax(dim="time")]
            min_var = ds[var].min(dim="time")
            min_time = ds[var].time[ds[var].argmin(dim="time")]
            result[f"max_{abbrev}"] = max_var
            result[f"max_{abbrev}_time"] = max_time
            result[f"min_{abbrev}"] = min_var
            result[f"min_{abbrev}_time"] = min_time
        return DataFrame(result)

    def _reference_lines_points(
        self,
        reftype: str = "lines",
        include_output: bool = True,
        datetime_to_str: bool = False,
    ) -> GeoDataFrame:
        if reftype == "lines":
            abbrev = "refln"
            gdf = super().reference_lines()
        elif reftype == "points":
            abbrev = "refpt"
            gdf = super().reference_points()
        else:
            raise ValueError('reftype must be either "lines" or "points".')
        if include_output is False:
            return gdf
        summary_output = self.reference_summary_output(reftype=reftype)
        gdf = gdf.merge(
            summary_output,
            on=[f"{abbrev}_id", f"{abbrev}_name", "mesh_name"],
            how="left",
        )
        if datetime_to_str:
            gdf = df_datetimes_to_str(gdf)
        return gdf

    def reference_lines(
        self, include_output: bool = True, datetime_to_str: bool = False
    ) -> GeoDataFrame:
        """Return the reference lines from a HEC-RAS HDF plan file.

        Includes summary output data for each reference line:
        - Maximum flow & time (max_q, max_q_time)
        - Minimum flow & time (min_q, min_q_time)
        - Maximum water surface elevation & time (max_ws, max_ws_time)
        - Minimum water surface elevation & time (min_ws, min_ws_time)

        Parameters
        ----------
        include_output : bool, optional
            If True, include summary output data in the GeoDataFrame. (default: True)
        datetime_to_str : bool, optional
            If True, convert datetime columns to strings. (default: False)

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame with reference line geometry and summary output data.
        """
        return self._reference_lines_points(
            reftype="lines",
            include_output=include_output,
            datetime_to_str=datetime_to_str,
        )

    def reference_points(
        self, include_output: bool = True, datetime_to_str: bool = False
    ) -> GeoDataFrame:
        """Return the reference points from a HEC-RAS HDF plan file.

        Parameters
        ----------
        include_output : bool, optional
            If True, include summary output data in the GeoDataFrame. (default: True)
        datetime_to_str : bool, optional
            If True, convert datetime columns to strings. (default: False)

        Includes summary output data for each reference point:
        - Maximum flow & time (max_q, max_q_time)
        - Minimum flow & time (min_q, min_q_time)
        - Maximum water surface elevation & time (max_ws, max_ws_time)
        - Minimum water surface elevation & time (min_ws, min_ws_time)

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame with reference point geometry and summary output data.
        """
        return self._reference_lines_points(
            reftype="points",
            include_output=include_output,
            datetime_to_str=datetime_to_str,
        )

    def get_plan_info_attrs(self) -> Dict:
        """Return plan information attributes from a HEC-RAS HDF plan file.

        Returns
        -------
        dict
            Dictionary of plan information attributes.
        """
        return self.get_attrs(self.PLAN_INFO_PATH)

    def get_plan_param_attrs(self) -> Dict:
        """Return plan parameter attributes from a HEC-RAS HDF plan file.

        Returns
        -------
        dict
            Dictionary of plan parameter attributes.
        """
        return self.get_attrs(self.PLAN_PARAMS_PATH)

    def get_meteorology_precip_attrs(self) -> Dict:
        """Return precipitation attributes from a HEC-RAS HDF plan file.

        Returns
        -------
        dict
            Dictionary of precipitation attributes.
        """
        return self.get_attrs(self.PRECIP_PATH)

    def get_results_unsteady_attrs(self) -> Dict:
        """Return unsteady attributes from a HEC-RAS HDF plan file.

        Returns
        -------
        dict
            Dictionary of unsteady attributes.
        """
        return self.get_attrs(self.RESULTS_UNSTEADY_PATH)

    def get_results_unsteady_summary_attrs(self) -> Dict:
        """Return results unsteady summary attributes from a HEC-RAS HDF plan file.

        Returns
        -------
        dict
            Dictionary of results summary attributes.
        """
        return self.get_attrs(self.RESULTS_UNSTEADY_SUMMARY_PATH)

    def get_results_volume_accounting_attrs(self) -> Dict:
        """Return volume accounting attributes from a HEC-RAS HDF plan file.

        Returns
        -------
        dict
            Dictionary of volume accounting attributes.
        """
        return self.get_attrs(self.VOLUME_ACCOUNTING_PATH)

    def enroachment_points(self) -> GeoDataFrame:  # noqa: D102
        raise NotImplementedError

    def steady_flow_names(self) -> list:
        """Return the profile information for each steady flow event.

        Returns
        -------
        DataFrame
            A Dataframe containing the profile names for each event
        """
        if self.STEADY_PROFILES_PATH not in self:
            return pd.DataFrame()

        profile_data = self[self.STEADY_PROFILES_PATH]
        profile_attrs = profile_data["Profile Names"][()]

        return [x.decode("utf-8") for x in profile_attrs]

    def steady_profile_xs_output(
        self, var: XsSteadyOutputVar, round_to: int = 2
    ) -> DataFrame:
        """Create a Dataframe from steady cross section results based on path.

        Parameters
        ----------
        var : XsSteadyOutputVar
            The name of the table in the steady results that information is to be retireved from.

        round_to : int, optional
            Number of decimal places to round output data to.

        Returns
        -------
            Dataframe with desired hdf data.
        """
        if var in XS_STEADY_OUTPUT_ADDITIONAL:
            path = f"{self.STEADY_XS_ADDITIONAL_PATH}/{var.value}"
        else:
            path = f"{self.STEADY_XS_PATH}/{var.value}"
        if path not in self:
            return DataFrame()

        profiles = self.steady_flow_names()

        steady_data = self[path]
        df = DataFrame(steady_data, index=profiles)
        df_t = df.T.copy()
        for p in profiles:
            df_t[p] = df_t[p].apply(lambda x: round(x, round_to))

        return df_t

    def cross_sections_wsel(self) -> DataFrame:
        """Return the water surface elevation information for each 1D Cross Section.

        Returns
        -------
        DataFrame
            A Dataframe containing the water surface elevations for each cross section and event
        """
        return self.steady_profile_xs_output(XsSteadyOutputVar.WATER_SURFACE)

    def cross_sections_flow(self) -> DataFrame:
        """Return the Flow information for each 1D Cross Section.

        Returns
        -------
        DataFrame
            A Dataframe containing the flow for each cross section and event
        """
        return self.steady_profile_xs_output(XsSteadyOutputVar.FLOW)

    def cross_sections_energy_grade(self) -> DataFrame:
        """Return the energy grade information for each 1D Cross Section.

        Returns
        -------
        DataFrame
            A Dataframe containing the water surface elevations for each cross section and event
        """
        return self.steady_profile_xs_output(XsSteadyOutputVar.ENERGY_GRADE)

    def cross_sections_additional_enc_station_left(self) -> DataFrame:
        """Return the left side encroachment information for a floodway plan hdf.

        Returns
        -------
        DataFrame
            A DataFrame containing the cross sections left side encroachment stations
        """
        return self.steady_profile_xs_output(
            XsSteadyOutputVar.ENCROACHMENT_STATION_LEFT
        )

    def cross_sections_additional_enc_station_right(self) -> DataFrame:
        """Return the right side encroachment information for a floodway plan hdf.

        Returns
        -------
        DataFrame
            A DataFrame containing the cross sections right side encroachment stations
        """
        return self.steady_profile_xs_output(
            XsSteadyOutputVar.ENCROACHMENT_STATION_RIGHT
        )

    def cross_sections_additional_area_total(self) -> DataFrame:
        """Return the 1D cross section area for each profile.

        Returns
        -------
        DataFrame
            A DataFrame containing the wet area inside the cross sections
        """
        return self.steady_profile_xs_output(XsSteadyOutputVar.AREA_INEFFECTIVE_TOTAL)

    def cross_sections_additional_velocity_total(self) -> DataFrame:
        """Return the 1D cross section velocity for each profile.

        Returns
        -------
        DataFrame
            A DataFrame containing the velocity inside the cross sections
        """
        return self.steady_profile_xs_output(XsSteadyOutputVar.VELOCITY_TOTAL)

    def _zmeta(self, ds: xr.Dataset) -> Dict:
        """Given a xarray Dataset, return kerchunk-style zarr reference metadata."""
        from kerchunk.hdf import SingleHdf5ToZarr
        import zarr
        import base64

        encoding = {}
        chunk_meta = {}

        # Loop through each variable / DataArray in the Dataset
        for var, da in ds.data_vars.items():
            # The "hdf_path" attribute is the path within the HDF5 file
            # that the DataArray was read from. This is attribute is inserted
            # by rashdf (see "mesh_timeseries_output" method).
            hdf_ds_path = da.attrs["hdf_path"]
            hdf_ds = self.get(hdf_ds_path)
            if hdf_ds is None:
                # If we don't know where in the HDF5 the data came from, we
                # have to skip it, because we won't be able to generate the
                # correct metadata for it.
                continue
            # Get the filters and storage info for the HDF5 dataset.
            # Calling private methods from Kerchunk here because
            # there's not a nice public API for this part. This is hacky
            # and a bit risky because these private methods are more likely
            # to change, but short of reimplementing these functions ourselves
            # it's the best way to get the metadata we need.
            # TODO: raise an issue in Kerchunk to expose this functionality?
            filters = SingleHdf5ToZarr._decode_filters(None, hdf_ds)
            encoding[var] = {"compressor": None, "filters": filters}
            storage_info = SingleHdf5ToZarr._storage_info(None, hdf_ds)
            # Generate chunk metadata for the DataArray
            for key, value in storage_info.items():
                chunk_number = ".".join([str(k) for k in key])
                chunk_key = f"{var}/{chunk_number}"
                chunk_meta[chunk_key] = [str(self._loc), value["offset"], value["size"]]
        # "Write" the Dataset to a temporary in-memory zarr store (which
        # is the same a Python dictionary)
        zarr_tmp = zarr.MemoryStore()
        # Use compute=False here because we don't _actually_ want to write
        # the data to the zarr store, we just want to generate the metadata.
        ds.to_zarr(zarr_tmp, mode="w", compute=False, encoding=encoding)
        zarr_meta = {"version": 1, "refs": {}}
        # Loop through the in-memory Zarr store, decode the data to strings,
        # and add it to the final metadata dictionary.
        for key, value in zarr_tmp.items():
            try:
                value_str = value.decode("utf-8")
            except UnicodeDecodeError:
                value_str = "base64:" + base64.b64encode(value).decode("utf-8")
            zarr_meta["refs"][key] = value_str
        zarr_meta["refs"].update(chunk_meta)
        return zarr_meta

    def zmeta_mesh_cells_timeseries_output(self, mesh_name: str) -> Dict:
        """Return kerchunk-style zarr reference metadata.

        Requires the 'zarr' and 'kerchunk' packages.

        Returns
        -------
        dict
            Dictionary of kerchunk-style zarr reference metadata.
        """
        ds = self._mesh_timeseries_outputs(
            mesh_name, TIME_SERIES_OUTPUT_VARS_CELLS, truncate=False
        )
        return self._zmeta(ds)

    def zmeta_mesh_faces_timeseries_output(self, mesh_name: str) -> Dict:
        """Return kerchunk-style zarr reference metadata.

        Requires the 'zarr' and 'kerchunk' packages.

        Returns
        -------
        dict
            Dictionary of kerchunk-style zarr reference metadata.
        """
        ds = self._mesh_timeseries_outputs(
            mesh_name, TIME_SERIES_OUTPUT_VARS_FACES, truncate=False
        )
        return self._zmeta(ds)

    def zmeta_reference_lines_timeseries_output(self) -> Dict:
        """Return kerchunk-style zarr reference metadata.

        Requires the 'zarr' and 'kerchunk' packages.

        Returns
        -------
        dict
            Dictionary of kerchunk-style zarr reference metadata.
        """
        ds = self.reference_lines_timeseries_output()
        return self._zmeta(ds)

    def zmeta_reference_points_timeseries_output(self) -> Dict:
        """Return kerchunk-style zarr reference metadata.

        Requires the 'zarr' and 'kerchunk' packages.

        Returns
        -------
        dict
            Dictionary of kerchunk-style zarr reference metadata.
        """
        ds = self.reference_points_timeseries_output()
        return self._zmeta(ds)



from pathlib import Path
import h5py
from pyproj import CRS
from src.rashdf import RasGeomHdf

from . import _create_hdf_with_group_attrs, _gdf_matches_json, _gdf_matches_json_alt

TEST_DATA = Path("./tests/data")
MUNCIE_G05 = TEST_DATA / "ras/Muncie.g05.hdf"
COAL_G01 = TEST_DATA / "ras/Coal.g01.hdf"
BAXTER_P01 = TEST_DATA / "ras_1d/Baxter.p01.hdf"
TEST_JSON = TEST_DATA / "json"
BALD_EAGLE_P18_REF = TEST_DATA / "ras/BaldEagleDamBrk.reflines-refpts.p18.hdf"

TEST_ATTRS = {"test_attribute1": "test_str1", "test_attribute2": 500}


def test_projection(tmp_path):
    wkt = 'PROJCS["Albers_Conic_Equal_Area",GEOGCS["GCS_North_American_1983",DATUM["D_North_American_1983",SPHEROID["GRS_1980",6378137.0,298.257222101]],PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Albers"],PARAMETER["false_easting",0.0],PARAMETER["false_northing",0.0],PARAMETER["central_meridian",-96.0],PARAMETER["standard_parallel_1",29.5],PARAMETER["standard_parallel_2",45.5],PARAMETER["latitude_of_origin",37.5],UNIT["Foot_US",0.3048006096012192]]'
    # Create a dummy HDF file
    with h5py.File(tmp_path / "test.hdf", "w") as f:
        f.attrs["Projection"] = wkt.encode("utf-8")
    # Open the HDF file
    ras_hdf = RasGeomHdf(tmp_path / "test.hdf")
    # Test the projection
    assert ras_hdf.projection() == CRS.from_wkt(wkt)


def test_mesh_area_names():
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert ghdf.mesh_area_names() == ["2D Interior Area", "Perimeter_NW"]


def test_mesh_areas():
    mesh_areas_json = TEST_JSON / "mesh_areas.json"
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert _gdf_matches_json(ghdf.mesh_areas(), mesh_areas_json)


def test_mesh_cell_faces():
    mesh_cell_faces_json = TEST_JSON / "mesh_cell_faces.json"
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert _gdf_matches_json(ghdf.mesh_cell_faces(), mesh_cell_faces_json)


def test_mesh_cell_points():
    mesh_cell_points_json = TEST_JSON / "mesh_cell_points.json"
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert _gdf_matches_json(ghdf.mesh_cell_points(), mesh_cell_points_json)


def test_mesh_cell_polygons():
    mesh_cell_polygons_json = TEST_JSON / "mesh_cell_polygons.json"
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert _gdf_matches_json(ghdf.mesh_cell_polygons(), mesh_cell_polygons_json)


def test_mesh_cell_polygons_coal():
    """Test with the mesh from the Coal River model.

    The Jan 2024 Coal River model from the Kanawha FFRD pilot project
    contains some topologically incorrect polygons in the 2D mesh;
    some of the mesh cell faces overlap.

    See: https://github.com/fema-ffrd/rashdf/issues/31
    """
    coal_bad_polygons_json = TEST_JSON / "coal-bad-mesh-cell-polygons.json"
    with RasGeomHdf(COAL_G01) as geom_hdf:
        gdf = geom_hdf.mesh_cell_polygons()
        gdf_bad_polygons = gdf[gdf["cell_id"].isin([8561, 11791, 17529])].to_crs(
            "EPSG:4326"
        )  # reproject because the Kanawha FFRD pilot project uses a custom CRS
        assert _gdf_matches_json(gdf_bad_polygons, coal_bad_polygons_json)


def test_bc_lines():
    bc_lines_json = TEST_JSON / "bc_lines.json"
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert _gdf_matches_json(ghdf.bc_lines(), bc_lines_json)


def test_breaklines():
    breaklines_json = TEST_JSON / "breaklines.json"
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert _gdf_matches_json(ghdf.breaklines(), breaklines_json)


def test_refinement_regions():
    rr_json = TEST_JSON / "refinement_regions.json"
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert _gdf_matches_json(ghdf.refinement_regions(), rr_json)


def test_get_geom_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(test_hdf, RasGeomHdf.GEOM_PATH, TEST_ATTRS)
    ras_hdf = RasGeomHdf(test_hdf)
    assert ras_hdf.get_geom_attrs() == TEST_ATTRS


def test_get_geom_structures_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(test_hdf, RasGeomHdf.GEOM_STRUCTURES_PATH, TEST_ATTRS)
    ras_hdf = RasGeomHdf(test_hdf)
    assert ras_hdf.get_geom_structures_attrs() == TEST_ATTRS


def test_get_geom_2d_flow_area_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(
        test_hdf, f"{RasGeomHdf.FLOW_AREA_2D_PATH}/group", TEST_ATTRS
    )
    ras_hdf = RasGeomHdf(test_hdf)
    assert ras_hdf.get_geom_2d_flow_area_attrs() == TEST_ATTRS


def test_structs():
    structs_json = TEST_JSON / "structures.json"
    with RasGeomHdf(MUNCIE_G05) as ghdf:
        assert _gdf_matches_json(ghdf.structures(datetime_to_str=True), structs_json)


def test_reference_lines_names():
    with RasGeomHdf(BALD_EAGLE_P18_REF) as geom_hdf:
        assert geom_hdf.reference_lines_names() == {
            "BaldEagleCr": [
                "Reference Line 1",
                "Reference Line 2",
                "Reference Line 3",
                "Reference Line 4",
            ]
        }
        assert geom_hdf.reference_lines_names("Upper 2D Area") == []


def test_reference_points_names():
    with RasGeomHdf(BALD_EAGLE_P18_REF) as geom_hdf:
        assert geom_hdf.reference_points_names() == {
            "Upper 2D Area": [
                "Reference Point 1",
            ],
            "BaldEagleCr": [
                "Reference Point 2",
                "Reference Point 3",
            ],
        }
        assert geom_hdf.reference_points_names("Upper 2D Area") == [
            "Reference Point 1",
        ]


def test_structs_not_found():
    with RasGeomHdf(COAL_G01) as ghdf:
        assert (ghdf.structures(), None)


def test_cross_sections():
    cross_section_json = TEST_JSON / "cross_sections.json"
    with RasGeomHdf(BAXTER_P01) as ghdf:
        assert _gdf_matches_json_alt(
            ghdf.cross_sections(datetime_to_str=True), cross_section_json
        )


def test_cross_sections_not_found():
    with RasGeomHdf(COAL_G01) as ghdf:
        assert (ghdf.cross_sections(), None)


def test_river_reaches():
    river_reaches_json = TEST_JSON / "river_reaches.json"
    with RasGeomHdf(BAXTER_P01) as ghdf:
        assert _gdf_matches_json_alt(ghdf.river_reaches(), river_reaches_json)


def test_river_reaches_not_found():
    with RasGeomHdf(COAL_G01) as ghdf:
        assert (ghdf.river_reaches(), None)


def test_cross_sections_elevations():
    xs_elevs_json = TEST_JSON / "xs_elevations.json"
    with RasGeomHdf(BAXTER_P01) as ghdf:
        assert _gdf_matches_json_alt(ghdf.cross_sections_elevations(), xs_elevs_json)


def test_cross_sections_elevations_not_found():
    with RasGeomHdf(COAL_G01) as ghdf:
        assert (ghdf.cross_sections_elevations(), None)



from src.rashdf.base import RasHdf
from unittest.mock import patch


def test_open():
    rasfile = "Muncie.g05.hdf"
    rasfile_path = f"./tests/data/ras/{rasfile}"
    hdf = RasHdf(rasfile_path)
    assert hdf._loc == rasfile_path


def test_open_uri():
    rasfile = "Muncie.g05.hdf"
    rasfile_path = f"./tests/data/ras/{rasfile}"
    url = f"s3://mybucket/{rasfile}"

    # Mock the specific functions used by s3fs
    with patch("s3fs.core.S3FileSystem.open", return_value=open(rasfile_path, "rb")):
        hdf = RasHdf.open_uri(url)
        assert hdf._loc == url



"""HEC-RAS Geometry HDF class."""

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame
from pyproj import CRS
from shapely import (
    Geometry,
    Polygon,
    Point,
    LineString,
    MultiLineString,
    MultiPolygon,
    Point,
    Polygon,
    polygonize_full,
)

from typing import Dict, List, Optional, Union


from .base import RasHdf
from .utils import (
    convert_ras_hdf_string,
    convert_ras_hdf_value,
    get_first_hdf_group,
    hdf5_attrs_to_dict,
)


class RasGeomHdfError(Exception):
    """HEC-RAS Plan HDF error class."""

    pass


class RasGeomHdf(RasHdf):
    """HEC-RAS Geometry HDF class."""

    GEOM_PATH = "Geometry"
    GEOM_STRUCTURES_PATH = f"{GEOM_PATH}/Structures"
    FLOW_AREA_2D_PATH = f"{GEOM_PATH}/2D Flow Areas"
    BC_LINES_PATH = f"{GEOM_PATH}/Boundary Condition Lines"
    BREAKLINES_PATH = f"{GEOM_PATH}/2D Flow Area Break Lines"
    REFERENCE_LINES_PATH = f"{GEOM_PATH}/Reference Lines"
    REFERENCE_POINTS_PATH = f"{GEOM_PATH}/Reference Points"
    CROSS_SECTIONS = f"{GEOM_PATH}/Cross Sections"
    RIVER_CENTERLINES = f"{GEOM_PATH}/River Centerlines"

    def __init__(self, name: str, **kwargs):
        """Open a HEC-RAS Geometry HDF file.

        Parameters
        ----------
        name : str
            The path to the RAS Geometry HDF file.
        kwargs : dict
            Additional keyword arguments to pass to h5py.File
        """
        super().__init__(name, **kwargs)

    def projection(self) -> Optional[CRS]:
        """Return the projection of the RAS geometry as a pyproj.CRS object.

        Returns
        -------
        pyproj.CRS or None
            The projection of the RAS geometry.
        """
        proj_wkt = self.attrs.get("Projection")
        if proj_wkt is None:
            return None
        if isinstance(proj_wkt, bytes) or isinstance(proj_wkt, np.bytes_):
            proj_wkt = proj_wkt.decode("utf-8")
        return CRS.from_wkt(proj_wkt)

    def mesh_area_names(self) -> List[str]:
        """Return a list of the 2D mesh area names of the RAS geometry.

        Returns
        -------
        List[str]
            A list of the 2D mesh area names (str) within the RAS geometry if 2D areas exist.
        """
        if self.FLOW_AREA_2D_PATH not in self:
            return list()
        return list(
            [
                convert_ras_hdf_string(n)
                for n in self[f"{self.FLOW_AREA_2D_PATH}/Attributes"][()]["Name"]
            ]
        )

    def mesh_areas(self) -> GeoDataFrame:
        """Return 2D flow area perimeter polygons.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the 2D flow area perimeter polygons if 2D areas exist.
        """
        mesh_area_names = self.mesh_area_names()
        if not mesh_area_names:
            return GeoDataFrame()
        mesh_area_polygons = [
            Polygon(self[f"{self.FLOW_AREA_2D_PATH}/{n}/Perimeter"][()])
            for n in mesh_area_names
        ]
        return GeoDataFrame(
            {"mesh_name": mesh_area_names, "geometry": mesh_area_polygons},
            geometry="geometry",
            crs=self.projection(),
        )

    def mesh_cell_polygons(self) -> GeoDataFrame:
        """Return 2D flow mesh cell polygons.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the 2D flow mesh cell polygons if 2D areas exist.
        """
        mesh_area_names = self.mesh_area_names()
        if not mesh_area_names:
            return GeoDataFrame()

        face_gdf = self.mesh_cell_faces()

        cell_dict = {"mesh_name": [], "cell_id": [], "geometry": []}
        for i, mesh_name in enumerate(mesh_area_names):
            cell_cnt = self[f"{self.FLOW_AREA_2D_PATH}/Cell Info"][()][i][1]
            cell_ids = list(range(cell_cnt))
            cell_face_info = self[
                f"{self.FLOW_AREA_2D_PATH}/{mesh_name}/Cells Face and Orientation Info"
            ][()]
            cell_face_values = self[
                f"{self.FLOW_AREA_2D_PATH}/{mesh_name}/Cells Face and Orientation Values"
            ][()][:, 0]
            face_id_lists = list(
                np.vectorize(
                    lambda cell_id: str(
                        cell_face_values[
                            cell_face_info[cell_id][0] : cell_face_info[cell_id][0]
                            + cell_face_info[cell_id][1]
                        ]
                    )
                )(cell_ids)
            )
            mesh_faces = (
                face_gdf[face_gdf.mesh_name == mesh_name][["face_id", "geometry"]]
                .set_index("face_id")
                .to_numpy()
            )
            cell_dict["mesh_name"] += [mesh_name] * cell_cnt
            cell_dict["cell_id"] += cell_ids
            cell_dict["geometry"] += list(
                np.vectorize(
                    lambda face_id_list: (
                        lambda geom_col: Polygon((geom_col[0] or geom_col[3]).geoms[0])
                    )(
                        polygonize_full(
                            np.ravel(
                                mesh_faces[
                                    np.array(face_id_list.strip("[]").split()).astype(
                                        int
                                    )
                                ]
                            )
                        )
                    )
                )(face_id_lists)
            )
        return GeoDataFrame(cell_dict, geometry="geometry", crs=self.projection())

    def mesh_cell_points(self) -> GeoDataFrame:
        """Return 2D flow mesh cell points.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the 2D flow mesh cell points if 2D areas exist.
        """
        mesh_area_names = self.mesh_area_names()
        if not mesh_area_names:
            return GeoDataFrame()
        pnt_dict = {"mesh_name": [], "cell_id": [], "geometry": []}
        for i, mesh_name in enumerate(mesh_area_names):
            starting_row, count = self[f"{self.FLOW_AREA_2D_PATH}/Cell Info"][()][i]
            cell_pnt_coords = self[f"{self.FLOW_AREA_2D_PATH}/Cell Points"][()][
                starting_row : starting_row + count
            ]
            pnt_dict["mesh_name"] += [mesh_name] * cell_pnt_coords.shape[0]
            pnt_dict["cell_id"] += range(count)
            pnt_dict["geometry"] += list(
                np.vectorize(lambda coords: Point(coords), signature="(n)->()")(
                    cell_pnt_coords
                )
            )
        return GeoDataFrame(pnt_dict, geometry="geometry", crs=self.projection())

    def mesh_cell_faces(self) -> GeoDataFrame:
        """Return 2D flow mesh cell faces.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the 2D flow mesh cell faces if 2D areas exist.
        """
        mesh_area_names = self.mesh_area_names()
        if not mesh_area_names:
            return GeoDataFrame()
        face_dict = {"mesh_name": [], "face_id": [], "geometry": []}
        for mesh_name in mesh_area_names:
            facepoints_index = self[
                f"{self.FLOW_AREA_2D_PATH}/{mesh_name}/Faces FacePoint Indexes"
            ][()]
            facepoints_coordinates = self[
                f"{self.FLOW_AREA_2D_PATH}/{mesh_name}/FacePoints Coordinate"
            ][()]
            faces_perimeter_info = self[
                f"{self.FLOW_AREA_2D_PATH}/{mesh_name}/Faces Perimeter Info"
            ][()]
            faces_perimeter_values = self[
                f"{self.FLOW_AREA_2D_PATH}/{mesh_name}/Faces Perimeter Values"
            ][()]
            face_id = -1
            for pnt_a_index, pnt_b_index in facepoints_index:
                face_id += 1
                face_dict["mesh_name"].append(mesh_name)
                face_dict["face_id"].append(face_id)
                coordinates = list()
                coordinates.append(facepoints_coordinates[pnt_a_index])
                starting_row, count = faces_perimeter_info[face_id]
                if count > 0:
                    coordinates += list(
                        faces_perimeter_values[starting_row : starting_row + count]
                    )
                coordinates.append(facepoints_coordinates[pnt_b_index])
                face_dict["geometry"].append(LineString(coordinates))
        return GeoDataFrame(face_dict, geometry="geometry", crs=self.projection())

    def get_geom_attrs(self) -> Dict:
        """Return base geometry attributes from a HEC-RAS HDF file.

        Returns
        -------
        dict
            Dictionary filled with base geometry attributes.
        """
        return self.get_attrs(self.GEOM_PATH)

    def get_geom_structures_attrs(self) -> Dict:
        """Return geometry structures attributes from a HEC-RAS HDF file.

        Returns
        -------
        dict
            Dictionary filled with geometry structures attributes.
        """
        return self.get_attrs(self.GEOM_STRUCTURES_PATH)

    def get_geom_2d_flow_area_attrs(self) -> Dict:
        """Return geometry 2d flow area attributes from a HEC-RAS HDF file.

        Returns
        -------
        dict
            Dictionary filled with geometry 2d flow area attributes.
        """
        try:
            d2_flow_area = get_first_hdf_group(self.get(self.FLOW_AREA_2D_PATH))
        except AttributeError:
            raise AttributeError(
                f"Unable to get 2D Flow Area; {self.FLOW_AREA_2D_PATH} group not found in HDF5 file."
            )

        d2_flow_area_attrs = hdf5_attrs_to_dict(d2_flow_area.attrs)

        return d2_flow_area_attrs

    def _get_polylines(
        self,
        path: str,
        info_name: str = "Polyline Info",
        parts_name: str = "Polyline Parts",
        points_name: str = "Polyline Points",
    ) -> List[Geometry]:
        polyline_info_path = f"{path}/{info_name}"
        polyline_parts_path = f"{path}/{parts_name}"
        polyline_points_path = f"{path}/{points_name}"

        polyline_info = self[polyline_info_path][()]
        polyline_parts = self[polyline_parts_path][()]
        polyline_points = self[polyline_points_path][()]

        geoms = []
        for pnt_start, pnt_cnt, part_start, part_cnt in polyline_info:
            points = polyline_points[pnt_start : pnt_start + pnt_cnt]
            if part_cnt == 1:
                geoms.append(LineString(points))
            else:
                parts = polyline_parts[part_start : part_start + part_cnt]
                geoms.append(
                    MultiLineString(
                        list(
                            points[part_pnt_start : part_pnt_start + part_pnt_cnt]
                            for part_pnt_start, part_pnt_cnt in parts
                        )
                    )
                )
        return geoms

    def bc_lines(self) -> GeoDataFrame:
        """Return 2D mesh area boundary condition lines.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the 2D mesh area boundary condition lines if they exist.
        """
        if self.BC_LINES_PATH not in self:
            return GeoDataFrame()
        bc_line_data = self[self.BC_LINES_PATH]
        bc_line_ids = range(bc_line_data["Attributes"][()].shape[0])
        v_conv_str = np.vectorize(convert_ras_hdf_string)
        names = v_conv_str(bc_line_data["Attributes"][()]["Name"])
        mesh_names = v_conv_str(bc_line_data["Attributes"][()]["SA-2D"])
        types = v_conv_str(bc_line_data["Attributes"][()]["Type"])
        geoms = self._get_polylines(self.BC_LINES_PATH)
        return GeoDataFrame(
            {
                "bc_line_id": bc_line_ids,
                "name": names,
                "mesh_name": mesh_names,
                "type": types,
                "geometry": geoms,
            },
            geometry="geometry",
            crs=self.projection(),
        )

    def breaklines(self) -> GeoDataFrame:
        """Return 2D mesh area breaklines.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the 2D mesh area breaklines if they exist.
        """
        if self.BREAKLINES_PATH not in self:
            return GeoDataFrame()
        bl_line_data = self[self.BREAKLINES_PATH]
        bl_line_ids = range(bl_line_data["Attributes"][()].shape[0])
        names = np.vectorize(convert_ras_hdf_string)(
            bl_line_data["Attributes"][()]["Name"]
        )
        geoms = self._get_polylines(self.BREAKLINES_PATH)
        return GeoDataFrame(
            {"bl_id": bl_line_ids, "name": names, "geometry": geoms},
            geometry="geometry",
            crs=self.projection(),
        )

    def refinement_regions(self) -> GeoDataFrame:
        """Return 2D mesh area refinement regions.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the 2D mesh area refinement regions if they exist.
        """
        if "/Geometry/2D Flow Area Refinement Regions" not in self:
            return GeoDataFrame()
        rr_data = self["/Geometry/2D Flow Area Refinement Regions"]
        rr_ids = range(rr_data["Attributes"][()].shape[0])
        names = np.vectorize(convert_ras_hdf_string)(rr_data["Attributes"][()]["Name"])
        geoms = list()
        for pnt_start, pnt_cnt, part_start, part_cnt in rr_data["Polygon Info"][()]:
            points = rr_data["Polygon Points"][()][pnt_start : pnt_start + pnt_cnt]
            if part_cnt == 1:
                geoms.append(Polygon(points))
            else:
                parts = rr_data["Polygon Parts"][()][part_start : part_start + part_cnt]
                geoms.append(
                    MultiPolygon(
                        list(
                            points[part_pnt_start : part_pnt_start + part_pnt_cnt]
                            for part_pnt_start, part_pnt_cnt in parts
                        )
                    )
                )
        return GeoDataFrame(
            {"rr_id": rr_ids, "name": names, "geometry": geoms},
            geometry="geometry",
            crs=self.projection(),
        )

    def structures(self, datetime_to_str: bool = False) -> GeoDataFrame:
        """Return the model structures.

        Parameters
        ----------
        datetime_to_str : bool, optional
            If True, convert datetime values to string format (default: False).

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the model structures if they exist.
        """
        if self.GEOM_STRUCTURES_PATH not in self:
            return GeoDataFrame()
        struct_data = self[self.GEOM_STRUCTURES_PATH]
        v_conv_val = np.vectorize(convert_ras_hdf_value)
        sd_attrs = struct_data["Attributes"][()]
        struct_dict = {"struct_id": range(sd_attrs.shape[0])}
        struct_dict.update(
            {name: v_conv_val(sd_attrs[name]) for name in sd_attrs.dtype.names}
        )
        geoms = self._get_polylines(
            self.GEOM_STRUCTURES_PATH,
            info_name="Centerline Info",
            parts_name="Centerline Parts",
            points_name="Centerline Points",
        )
        struct_gdf = GeoDataFrame(
            struct_dict,
            geometry=geoms,
            crs=self.projection(),
        )
        if datetime_to_str:
            struct_gdf["Last Edited"] = struct_gdf["Last Edited"].apply(
                lambda x: pd.Timestamp.isoformat(x)
            )
        return struct_gdf

    def connections(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def ic_points(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def _reference_lines_points_names(
        self, reftype: str = "lines", mesh_name: Optional[str] = None
    ) -> Union[Dict[str, List[str]], List[str]]:
        """Return reference line names.

        If a mesh name is provided, return a list of the reference line names for that mesh area.
        If no mesh name is provided, return a dictionary of mesh names and their reference line names.

        Parameters
        ----------
        mesh_name : str, optional
            The name of the mesh area for which to return reference line names.

        Returns
        -------
        Union[Dict[str, List[str]], List[str]]
            A dictionary of mesh names and their reference line names if mesh_name is None.
            A list of reference line names for the specified mesh area if mesh_name is not None.
        """
        if reftype == "lines":
            path = self.REFERENCE_LINES_PATH
            sa_2d_field = "SA-2D"
        elif reftype == "points":
            path = self.REFERENCE_POINTS_PATH
            sa_2d_field = "SA/2D"
        else:
            raise RasGeomHdfError(
                f"Invalid reference type: {reftype} -- must be 'lines' or 'points'."
            )
        attributes_path = f"{path}/Attributes"
        if mesh_name is None and attributes_path not in self:
            return {m: [] for m in self.mesh_area_names()}
        if mesh_name is not None and attributes_path not in self:
            return []
        attributes = self[attributes_path][()]
        v_conv_str = np.vectorize(convert_ras_hdf_string)
        names = np.vectorize(convert_ras_hdf_string)(attributes["Name"])
        if mesh_name is not None:
            return names[v_conv_str(attributes[sa_2d_field]) == mesh_name].tolist()
        mesh_names = np.vectorize(convert_ras_hdf_string)(attributes[sa_2d_field])
        return {m: names[mesh_names == m].tolist() for m in np.unique(mesh_names)}

    def reference_lines_names(
        self, mesh_name: Optional[str] = None
    ) -> Union[Dict[str, List[str]], List[str]]:
        """Return reference line names.

        If a mesh name is provided, return a list of the reference line names for that mesh area.
        If no mesh name is provided, return a dictionary of mesh names and their reference line names.

        Parameters
        ----------
        mesh_name : str, optional
            The name of the mesh area for which to return reference line names.

        Returns
        -------
        Union[Dict[str, List[str]], List[str]]
            A dictionary of mesh names and their reference line names if mesh_name is None.
            A list of reference line names for the specified mesh area if mesh_name is not None.
        """
        return self._reference_lines_points_names("lines", mesh_name)

    def reference_points_names(
        self, mesh_name: Optional[str] = None
    ) -> Union[Dict[str, List[str]], List[str]]:
        """Return reference point names.

        If a mesh name is provided, return a list of the reference point names for that mesh area.
        If no mesh name is provided, return a dictionary of mesh names and their reference point names.

        Parameters
        ----------
        mesh_name : str, optional
            The name of the mesh area for which to return reference point names.

        Returns
        -------
        Union[Dict[str, List[str]], List[str]]
            A dictionary of mesh names and their reference point names if mesh_name is None.
            A list of reference point names for the specified mesh area if mesh_name is not None.
        """
        return self._reference_lines_points_names("points", mesh_name)

    def reference_lines(self) -> GeoDataFrame:
        """Return the reference lines geometry and attributes.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the reference lines if they exist.
        """
        attributes_path = f"{self.REFERENCE_LINES_PATH}/Attributes"
        if attributes_path not in self:
            return GeoDataFrame()
        attributes = self[attributes_path][()]
        refline_ids = range(attributes.shape[0])
        v_conv_str = np.vectorize(convert_ras_hdf_string)
        names = v_conv_str(attributes["Name"])
        mesh_names = v_conv_str(attributes["SA-2D"])
        try:
            types = v_conv_str(attributes["Type"])
        except ValueError:
            # "Type" field doesn't exist -- observed in some RAS HDF files
            types = np.array([""] * attributes.shape[0])
        geoms = self._get_polylines(self.REFERENCE_LINES_PATH)
        return GeoDataFrame(
            {
                "refln_id": refline_ids,
                "refln_name": names,
                "mesh_name": mesh_names,
                "type": types,
                "geometry": geoms,
            },
            geometry="geometry",
            crs=self.projection(),
        )

    def reference_points(self) -> GeoDataFrame:
        """Return the reference points geometry and attributes.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the reference points if they exist.
        """
        attributes_path = f"{self.REFERENCE_POINTS_PATH}/Attributes"
        if attributes_path not in self:
            return GeoDataFrame()
        ref_points_group = self[self.REFERENCE_POINTS_PATH]
        attributes = ref_points_group["Attributes"][:]
        v_conv_str = np.vectorize(convert_ras_hdf_string)
        names = v_conv_str(attributes["Name"])
        mesh_names = v_conv_str(attributes["SA/2D"])
        cell_id = attributes["Cell Index"]
        points = ref_points_group["Points"][()]
        return GeoDataFrame(
            {
                "refpt_id": range(attributes.shape[0]),
                "refpt_name": names,
                "mesh_name": mesh_names,
                "cell_id": cell_id,
                "geometry": list(map(Point, points)),
            },
            geometry="geometry",
            crs=self.projection(),
        )

    def pump_stations(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def mannings_calibration_regions(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def classification_polygons(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def terrain_modifications(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def cross_sections(self, datetime_to_str: bool = False) -> GeoDataFrame:
        """Return the model 1D cross sections.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the model 1D cross sections if they exist.
        """
        if self.CROSS_SECTIONS not in self:
            return GeoDataFrame()

        xs_data = self[self.CROSS_SECTIONS]
        v_conv_val = np.vectorize(convert_ras_hdf_value)
        xs_attrs = xs_data["Attributes"][()]
        xs_dict = {"xs_id": range(xs_attrs.shape[0])}
        xs_dict.update(
            {name: v_conv_val(xs_attrs[name]) for name in xs_attrs.dtype.names}
        )
        geoms = self._get_polylines(self.CROSS_SECTIONS)
        xs_gdf = GeoDataFrame(
            xs_dict,
            geometry=geoms,
            crs=self.projection(),
        )
        if datetime_to_str:
            xs_gdf["Last Edited"] = xs_gdf["Last Edited"].apply(
                lambda x: pd.Timestamp.isoformat(x)
            )
        return xs_gdf

    def river_reaches(self, datetime_to_str: bool = False) -> GeoDataFrame:
        """Return the model 1D river reach lines.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame containing the model 1D river reach lines if they exist.
        """
        if self.RIVER_CENTERLINES not in self:
            return GeoDataFrame()

        river_data = self[self.RIVER_CENTERLINES]
        v_conv_val = np.vectorize(convert_ras_hdf_value)
        river_attrs = river_data["Attributes"][()]
        river_dict = {"river_id": range(river_attrs.shape[0])}
        river_dict.update(
            {name: v_conv_val(river_attrs[name]) for name in river_attrs.dtype.names}
        )
        geoms = list()
        geoms = self._get_polylines(self.RIVER_CENTERLINES)
        river_gdf = GeoDataFrame(
            river_dict,
            geometry=geoms,
            crs=self.projection(),
        )
        if datetime_to_str:
            river_gdf["Last Edited"] = river_gdf["Last Edited"].apply(
                lambda x: pd.Timestamp.isoformat(x)
            )
        return river_gdf

    def flowpaths(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def bank_points(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def bank_lines(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def ineffective_areas(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def ineffective_points(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def blocked_obstructions(self) -> GeoDataFrame:  # noqa D102
        raise NotImplementedError

    def cross_sections_elevations(self, round_to: int = 2) -> pd.DataFrame:
        """Return the model cross section elevation information.

        Returns
        -------
        DataFrame
            A DataFrame containing the model cross section elevation information if they exist.
        """
        path = "/Geometry/Cross Sections"
        if path not in self:
            return pd.DataFrame()

        xselev_data = self[path]
        xs_df = self.cross_sections()
        elevations = list()
        for part_start, part_cnt in xselev_data["Station Elevation Info"][()]:
            xzdata = xselev_data["Station Elevation Values"][()][
                part_start : part_start + part_cnt
            ]
            elevations.append(xzdata)

        xs_elev_df = xs_df[
            ["xs_id", "River", "Reach", "RS", "Left Bank", "Right Bank"]
        ].copy()
        xs_elev_df["Left Bank"] = xs_elev_df["Left Bank"].round(round_to).astype(str)
        xs_elev_df["Right Bank"] = xs_elev_df["Right Bank"].round(round_to).astype(str)
        xs_elev_df["elevation info"] = elevations

        return xs_elev_df



from src.cli import (
    parse_args,
    export,
    docstring_to_help,
    fiona_supported_drivers,
    pyogrio_supported_drivers,
)

import geopandas as gpd
from pyproj import CRS
import pytest

import builtins
import json
from pathlib import Path

TEST_DATA = Path("./tests/data")
MUNCIE_G05 = TEST_DATA / "ras/Muncie.g05.hdf"
BALD_EAGLE_P18 = TEST_DATA / "ras/BaldEagleDamBrk.p18.hdf"


def test_docstring_to_help():
    docstring = """This is a test docstring.
    This is not part of the help message.
    """
    assert docstring_to_help(docstring) == "This is a test docstring."

    docstring = """Return the something or other.
    Blah blah blah."""
    assert docstring_to_help(docstring) == "Export the something or other."

    docstring = None
    assert docstring_to_help(docstring) == ""


def test_fiona_supported_drivers():
    drivers = fiona_supported_drivers()
    assert "ESRI Shapefile" in drivers
    assert "GeoJSON" in drivers
    assert "GPKG" in drivers
    assert "MBTiles" not in drivers


def test_pyogrio_supported_drivers():
    drivers = pyogrio_supported_drivers()
    assert "ESRI Shapefile" in drivers
    assert "GeoJSON" in drivers
    assert "GPKG" in drivers
    assert "MBTiles" in drivers


def test_parse_args():
    args = parse_args(["structures", "test.hdf"])
    assert args.func == "structures"
    assert args.hdf_file == "test.hdf"
    assert args.output_file is None
    assert args.to_crs is None
    assert not args.parquet
    assert not args.feather
    assert args.kwargs is None

    args = parse_args(["mesh_areas", "test.hdf", "test.json"])
    assert args.func == "mesh_areas"
    assert args.hdf_file == "test.hdf"
    assert args.output_file == "test.json"
    assert args.to_crs is None
    assert not args.parquet
    assert not args.feather
    assert args.kwargs is None

    args = parse_args(
        [
            "mesh_areas",
            "test.hdf",
            "test.json",
            "--to-crs",
            "EPSG:4326",
            "--parquet",
            "--kwargs",
            '{"compression": "gzip"}',
        ]
    )
    assert args.func == "mesh_areas"
    assert args.hdf_file == "test.hdf"
    assert args.output_file == "test.json"
    assert args.to_crs == "EPSG:4326"
    assert args.parquet
    assert not args.feather
    assert args.kwargs == '{"compression": "gzip"}'

    args = parse_args(["--fiona-drivers"])
    assert args.fiona_drivers


def test_export(tmp_path: Path):
    args = parse_args(["structures", str(MUNCIE_G05)])
    exported = json.loads(export(args))
    gdf = gpd.GeoDataFrame.from_features(exported)
    assert len(gdf) == 3
    assert gdf["Last Edited"].to_list() == [
        "2024-04-15T15:21:34",
        "2024-04-15T15:21:48",
        "2024-04-15T15:26:15",
    ]

    test_json_path = tmp_path / "test.json"
    args = parse_args(["mesh_areas", str(MUNCIE_G05), str(test_json_path)])
    export(args)
    gdf = gpd.read_file(test_json_path)
    assert len(gdf) == 2

    test_parquet_path = tmp_path / "test.parquet"
    args = parse_args(
        [
            "mesh_cell_points",
            str(MUNCIE_G05),
            str(test_parquet_path),
            "--parquet",
            "--to-crs",
            "EPSG:4326",
        ]
    )
    export(args)
    gdf = gpd.read_parquet(test_parquet_path)
    assert len(gdf) == 5790
    assert gdf.crs == CRS.from_epsg(4326)


def test_export_plan_hdf():
    args = parse_args(["mesh_cell_points", str(BALD_EAGLE_P18)])
    exported = json.loads(export(args))
    gdf = gpd.GeoDataFrame.from_features(exported)
    assert len(gdf) == 4425


def test_fiona_missing(monkeypatch):
    # Test behavior when fiona isn't installed

    def mock_import(name, globals=None, locals=None, fromlist=(), level=0):
        if name == "fiona":
            raise ImportError("No module named 'fiona'")
        return real_import(name, globals, locals, fromlist, level)

    real_import = builtins.__import__

    # Replace the built-in __import__ function with our mock
    monkeypatch.setattr(builtins, "__import__", mock_import)

    # Verify that the --fiona-drivers argument is not available
    # when fiona is not installed
    args = parse_args(["structures", "fake_file.hdf"])
    assert not hasattr(args, "fiona_drivers")


def test_print_pyogrio_supported_drivers(capfd):
    export(parse_args(["--pyogrio-drivers"]))
    captured = capfd.readouterr()
    assert "ESRI Shapefile" in captured.out
    assert "GeoJSON" in captured.out
    assert "GPKG" in captured.out
    assert "MBTiles" in captured.out


def test_print_fiona_supported_drivers(capfd):
    export(parse_args(["--fiona-drivers"]))
    captured = capfd.readouterr()
    assert "ESRI Shapefile" in captured.out
    assert "GeoJSON" in captured.out
    assert "GPKG" in captured.out
    assert "MBTiles" not in captured.out



from src.rashdf.plan import (
    RasPlanHdf,
    SummaryOutputVar,
    RasPlanHdfError,
    TimeSeriesOutputVar,
)

import filecmp
import json
from pathlib import Path

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
import pytest
import xarray as xr

from . import (
    _create_hdf_with_group_attrs,
    _gdf_matches_json,
    get_sha1_hash,
    _gdf_matches_json_alt,
)

TEST_DATA = Path("./tests/data")
TEST_JSON = TEST_DATA / "json"
TEST_CSV = TEST_DATA / "csv"
TEST_ATTRS = {"test_attribute1": "test_str1", "test_attribute2": 500}
BALD_EAGLE_P18 = TEST_DATA / "ras/BaldEagleDamBrk.p18.hdf"
BALD_EAGLE_P18_TIMESERIES = TEST_DATA / "ras/BaldEagleDamBrk.p18.timeseries.hdf"
BALD_EAGLE_P18_REF = TEST_DATA / "ras/BaldEagleDamBrk.reflines-refpts.p18.hdf"
DENTON = TEST_DATA / "ras/Denton.hdf"
MUNCIE_G05 = TEST_DATA / "ras/Muncie.g05.hdf"
COAL_G01 = TEST_DATA / "ras/Coal.g01.hdf"
BAXTER_P01 = TEST_DATA / "ras_1d/Baxter.p01.hdf"
FLODENCR_P01 = TEST_DATA / "ras_1d/FLODENCR.p01.hdf"


def test_get_plan_info_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(test_hdf, RasPlanHdf.PLAN_INFO_PATH, TEST_ATTRS)
    ras_plan_hdf = RasPlanHdf(test_hdf)
    assert ras_plan_hdf.get_plan_info_attrs() == TEST_ATTRS


def test_get_plan_param_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(test_hdf, RasPlanHdf.PLAN_PARAMS_PATH, TEST_ATTRS)
    ras_plan_hdf = RasPlanHdf(test_hdf)
    assert ras_plan_hdf.get_plan_param_attrs() == TEST_ATTRS


def test_get_meteorology_precip_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(test_hdf, RasPlanHdf.PRECIP_PATH, TEST_ATTRS)
    ras_plan_hdf = RasPlanHdf(test_hdf)
    assert ras_plan_hdf.get_meteorology_precip_attrs() == TEST_ATTRS


def test_get_results_unsteady_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(test_hdf, RasPlanHdf.RESULTS_UNSTEADY_PATH, TEST_ATTRS)
    ras_plan_hdf = RasPlanHdf(test_hdf)
    assert ras_plan_hdf.get_results_unsteady_attrs() == TEST_ATTRS


def test_get_results_unsteady_summary_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(
        test_hdf, RasPlanHdf.RESULTS_UNSTEADY_SUMMARY_PATH, TEST_ATTRS
    )
    ras_plan_hdf = RasPlanHdf(test_hdf)
    assert ras_plan_hdf.get_results_unsteady_summary_attrs() == TEST_ATTRS


def test_get_results_volume_accounting_attrs(tmp_path):
    test_hdf = tmp_path / "test.hdf"
    _create_hdf_with_group_attrs(
        test_hdf, RasPlanHdf.VOLUME_ACCOUNTING_PATH, TEST_ATTRS
    )
    ras_plan_hdf = RasPlanHdf(test_hdf)
    assert ras_plan_hdf.get_results_volume_accounting_attrs() == TEST_ATTRS


def test__mesh_summary_output_group_null():
    """Test that an exception is raised if the specified mesh summary output data doesn't exist."""
    with RasPlanHdf(BALD_EAGLE_P18) as plan_hdf:
        with pytest.raises(RasPlanHdfError):
            plan_hdf._mesh_summary_output_group(
                "Nonexistent Mesh", SummaryOutputVar.MINIMUM_WATER_SURFACE
            )


def test_mesh_cell_points():
    # Test that the mesh cell points are returned with no other columns if
    # "include_output" is False.
    mesh_cell_points_json = TEST_JSON / "mesh_cell_points.json"
    with RasPlanHdf(MUNCIE_G05) as plan_hdf:
        assert _gdf_matches_json(
            plan_hdf.mesh_cell_points(include_output=False), mesh_cell_points_json
        )


def test_include_output_type_error():
    with RasPlanHdf(BALD_EAGLE_P18) as plan_hdf:
        with pytest.raises(ValueError):
            plan_hdf.mesh_cell_points(include_output="This should be a list instead")


def test_cells_or_values_error():
    """Test that an exception is raised if the cells_or_faces parameter is not 'cells' or 'faces'."""
    with RasPlanHdf(BALD_EAGLE_P18) as plan_hdf:
        with pytest.raises(ValueError):
            plan_hdf._mesh_summary_outputs_gdf(
                "mesh_cell_points", cells_or_faces="neither"
            )


def test_mesh_cell_points_all_outputs_columns():
    """Test that all summary output columns are returned if include_output is True."""
    plan_hdf = RasPlanHdf(BALD_EAGLE_P18)
    gdf = plan_hdf.mesh_cell_points(include_output=True)
    expected_columns = [
        "mesh_name",
        "cell_id",
        "geometry",
        "max_iter",
        "last_iter",
        "max_ws_err",
        "max_ws_err_time",
        "max_ws",
        "max_ws_time",
        "min_ws",
        "min_ws_time",
    ]
    assert list(gdf.columns) == expected_columns


def test_mesh_cell_points_with_output(tmp_path):
    plan_hdf = RasPlanHdf(BALD_EAGLE_P18)
    gdf = plan_hdf.mesh_cell_points(
        datetime_to_str=True,
        include_output=[
            SummaryOutputVar.CELL_MAXIMUM_WATER_SURFACE_ERROR,
            "Maximum Water Surface",
            "Minimum Water Surface",
        ],
    )
    temp_points = tmp_path / "temp-bald-eagle-mesh-cell-points.geojson"
    gdf = gdf.to_crs(4326)
    gdf.to_file(temp_points, engine="fiona")
    valid = get_sha1_hash(TEST_JSON / "bald-eagle-mesh-cell-points.geojson")
    test = get_sha1_hash(temp_points)
    assert valid == test


def test_mesh_cell_polygons_with_output(tmp_path):
    plan_hdf = RasPlanHdf(BALD_EAGLE_P18)
    gdf = plan_hdf.mesh_cell_polygons(
        datetime_to_str=True,
        include_output=[
            SummaryOutputVar.CELL_MAXIMUM_WATER_SURFACE_ERROR,
            "Maximum Water Surface",
            "Minimum Water Surface",
        ],
    )
    temp_polygons = tmp_path / "temp-bald-eagle-mesh-cell-polygons.geojson"
    gdf.to_crs(4326).to_file(temp_polygons, engine="fiona")
    valid = get_sha1_hash(TEST_JSON / "bald-eagle-mesh-cell-polygons.geojson")
    test = get_sha1_hash(temp_polygons)
    assert valid == test


def test_mesh_cell_faces_with_output(tmp_path):
    plan_hdf = RasPlanHdf(BALD_EAGLE_P18)
    gdf = plan_hdf.mesh_cell_faces(
        datetime_to_str=True,
        include_output=[
            "Minimum Face Velocity",
            SummaryOutputVar.MAXIMUM_FACE_VELOCITY,
        ],
    )
    temp_faces = tmp_path / "temp-bald-eagle-mesh-cell-faces.geojson"
    gdf.to_crs(4326).to_file(temp_faces, engine="fiona")
    valid = get_sha1_hash(TEST_JSON / "bald-eagle-mesh-cell-faces.geojson")
    test = get_sha1_hash(temp_faces)
    assert valid == test


def test_mesh_timeseries_output():
    with RasPlanHdf(BALD_EAGLE_P18_TIMESERIES) as plan_hdf:
        with pytest.raises(ValueError):
            plan_hdf.mesh_timeseries_output(
                "Fake Mesh", TimeSeriesOutputVar.WATER_SURFACE
            )
        with pytest.raises(ValueError):
            plan_hdf.mesh_timeseries_output("BaldEagleCr", "Fake Variable")


def test_mesh_cells_timeseries_output():
    with RasPlanHdf(BALD_EAGLE_P18_TIMESERIES) as plan_hdf:
        ds = plan_hdf.mesh_cells_timeseries_output("BaldEagleCr")
        assert "time" in ds.coords
        assert "cell_id" in ds.coords
        assert "Water Surface" in ds.variables
        ws = ds["Water Surface"]
        assert ws.shape == (37, 3359)
        assert ws.attrs["units"] == "ft"
        assert ws.attrs["mesh_name"] == "BaldEagleCr"
        df = ws.sel(cell_id=123).to_dataframe()
        valid_df = pd.read_csv(
            TEST_CSV / "BaldEagleDamBrk.BaldEagleCr.ws.123.csv",
            index_col="time",
            parse_dates=True,
            dtype={"Water Surface": np.float32},
        )
        assert_frame_equal(df, valid_df)

        ds = plan_hdf.mesh_cells_timeseries_output("Upper 2D Area")
        assert "time" in ds.coords
        assert "cell_id" in ds.coords
        assert "Water Surface" in ds.variables
        ws = ds["Water Surface"]
        assert ws.shape == (37, 1066)
        assert ws.attrs["units"] == "ft"
        assert ws.attrs["mesh_name"] == "Upper 2D Area"
        df = ws.sel(cell_id=456).to_dataframe()
        valid_df = pd.read_csv(
            TEST_CSV / "BaldEagleDamBrk.Upper2DArea.ws.456.csv",
            index_col="time",
            parse_dates=True,
            dtype={"Water Surface": np.float32},
        )
        assert_frame_equal(df, valid_df)


def test_mesh_timeseries_output_cells():
    with pytest.warns(DeprecationWarning):
        with RasPlanHdf(BALD_EAGLE_P18_TIMESERIES) as plan_hdf:
            plan_hdf.mesh_timeseries_output_cells("BaldEagleCr")


def test_mesh_faces_timeseries_output():
    with RasPlanHdf(BALD_EAGLE_P18_TIMESERIES) as plan_hdf:
        ds = plan_hdf.mesh_faces_timeseries_output("BaldEagleCr")
        assert "time" in ds.coords
        assert "face_id" in ds.coords
        assert "Face Velocity" in ds.variables
        v = ds["Face Velocity"]
        assert v.shape == (37, 7295)
        assert v.attrs["units"] == "ft/s"
        assert v.attrs["mesh_name"] == "BaldEagleCr"
        df = v.sel(face_id=678).to_dataframe()
        valid_df = pd.read_csv(
            TEST_CSV / "BaldEagleDamBrk.BaldEagleCr.v.678.csv",
            index_col="time",
            parse_dates=True,
            dtype={"Face Velocity": np.float32},
        )
        assert_frame_equal(df, valid_df)

        ds = plan_hdf.mesh_faces_timeseries_output("Upper 2D Area")
        assert "time" in ds.coords
        assert "face_id" in ds.coords
        assert "Face Velocity" in ds.variables
        v = ds["Face Velocity"]
        assert v.shape == (37, 2286)
        assert v.attrs["units"] == "ft/s"
        assert v.attrs["mesh_name"] == "Upper 2D Area"
        df = v.sel(face_id=567).to_dataframe()
        valid_df = pd.read_csv(
            TEST_CSV / "BaldEagleDamBrk.Upper2DArea.v.567.csv",
            index_col="time",
            parse_dates=True,
            dtype={"Face Velocity": np.float32},
        )
        assert_frame_equal(df, valid_df)


def test_mesh_timeseries_output_faces():
    with pytest.warns(DeprecationWarning):
        with RasPlanHdf(BALD_EAGLE_P18_TIMESERIES) as plan_hdf:
            plan_hdf.mesh_timeseries_output_faces("BaldEagleCr")


def test_reference_lines(tmp_path: Path):
    plan_hdf = RasPlanHdf(BALD_EAGLE_P18_REF)
    gdf = plan_hdf.reference_lines(datetime_to_str=True)
    temp_lines = tmp_path / "temp-bald-eagle-reference-lines.geojson"
    gdf.to_crs(4326).to_file(temp_lines, engine="fiona")
    with open(TEST_JSON / "bald-eagle-reflines.geojson") as f:
        valid_lines = f.read()
        with open(temp_lines) as f:
            test_lines = f.read()
            assert valid_lines == test_lines


def test_reference_lines_timeseries(tmp_path: Path):
    plan_hdf = RasPlanHdf(BALD_EAGLE_P18_REF)
    ds = plan_hdf.reference_lines_timeseries_output()
    assert "time" in ds.coords
    assert "refln_id" in ds.coords
    assert "refln_name" in ds.coords
    assert "mesh_name" in ds.coords
    assert "Water Surface" in ds.variables
    assert "Flow" in ds.variables

    ws = ds["Water Surface"]
    assert ws.shape == (37, 4)
    assert ws.attrs["units"] == "ft"
    q = ds["Flow"]
    assert q.shape == (37, 4)
    assert q.attrs["units"] == "cfs"

    df = ds.sel(refln_id=2).to_dataframe()
    valid_df = pd.read_csv(
        TEST_CSV / "BaldEagleDamBrk.reflines.2.csv",
        index_col="time",
        parse_dates=True,
        dtype={"Water Surface": np.float32, "Flow": np.float32},
    )
    assert_frame_equal(df, valid_df)


def test_reference_points(tmp_path: Path):
    plan_hdf = RasPlanHdf(BALD_EAGLE_P18_REF)
    gdf = plan_hdf.reference_points(datetime_to_str=True)
    temp_lines = tmp_path / "temp-bald-eagle-reference-points.geojson"
    gdf.to_crs(4326).to_file(temp_lines, engine="fiona")
    with open(TEST_JSON / "bald-eagle-refpoints.geojson") as f:
        valid_points = f.read()
        with open(temp_lines) as f:
            test_points = f.read()
            assert valid_points == test_points


def test_reference_points_timeseries():
    plan_hdf = RasPlanHdf(BALD_EAGLE_P18_REF)
    ds = plan_hdf.reference_points_timeseries_output()
    assert "time" in ds.coords
    assert "refpt_id" in ds.coords
    assert "refpt_name" in ds.coords
    assert "mesh_name" in ds.coords
    assert "Water Surface" in ds.variables
    assert "Velocity" in ds.variables

    ws = ds["Water Surface"]
    assert ws.shape == (37, 3)
    assert ws.attrs["units"] == "ft"
    v = ds["Velocity"]
    assert v.attrs["units"] == "ft/s"
    assert v.shape == (37, 3)

    df = ds.sel(refpt_id=1).to_dataframe()
    valid_df = pd.read_csv(
        TEST_CSV / "BaldEagleDamBrk.refpoints.1.csv",
        index_col="time",
        parse_dates=True,
        dtype={"Water Surface": np.float32, "Velocity": np.float32},
    )
    assert_frame_equal(df, valid_df)


def test_cross_sections_additional_velocity_total():
    xs_velocity_json = TEST_JSON / "xs_velocity.json"
    with RasPlanHdf(BAXTER_P01) as phdf:
        assert _gdf_matches_json_alt(
            phdf.cross_sections_additional_velocity_total(), xs_velocity_json
        )


def test_cross_sections_additional_velocity_total_not_found():
    with RasPlanHdf(COAL_G01) as phdf:
        assert (phdf.cross_sections_additional_velocity_total(), None)


def test_cross_sections_additional_area_total():
    xs_area_json = TEST_JSON / "xs_area.json"
    with RasPlanHdf(BAXTER_P01) as phdf:
        assert _gdf_matches_json_alt(
            phdf.cross_sections_additional_area_total(), xs_area_json
        )


def test_cross_sections_additional_area_total_not_found():
    with RasPlanHdf(COAL_G01) as phdf:
        assert (phdf.cross_sections_additional_area_total(), None)


def test_steady_flow_names():
    with RasPlanHdf(BAXTER_P01) as phdf:
        assert phdf.steady_flow_names() == ["Big"]


def test_steady_flow_names_not_found():
    with RasPlanHdf(COAL_G01) as phdf:
        assert (phdf.steady_flow_names(), None)


def test_cross_sections_wsel():
    xs_wsel_json = TEST_JSON / "xs_wsel.json"
    with RasPlanHdf(BAXTER_P01) as phdf:
        assert _gdf_matches_json_alt(phdf.cross_sections_wsel(), xs_wsel_json)


def test_cross_sections_wsel_not_found():
    with RasPlanHdf(COAL_G01) as phdf:
        assert (phdf.cross_sections_wsel(), None)


def test_cross_sections_additional_enc_station_right():
    xs_enc_station_right_json = TEST_JSON / "xs_enc_station_right.json"
    with RasPlanHdf(FLODENCR_P01) as phdf:
        assert _gdf_matches_json_alt(
            phdf.cross_sections_additional_enc_station_right(),
            xs_enc_station_right_json,
        )


def test_cross_sections_additional_enc_station_right_not_found():
    with RasPlanHdf(COAL_G01) as phdf:
        assert (phdf.cross_sections_additional_enc_station_right(), None)


def test_cross_sections_additional_enc_station_left():
    xs_enc_station_left_json = TEST_JSON / "xs_enc_station_left.json"
    with RasPlanHdf(FLODENCR_P01) as phdf:
        assert _gdf_matches_json_alt(
            phdf.cross_sections_additional_enc_station_left(), xs_enc_station_left_json
        )


def test_cross_sections_additional_enc_station_left_not_found():
    with RasPlanHdf(COAL_G01) as phdf:
        assert (phdf.cross_sections_additional_enc_station_left(), None)


def test_cross_sections_flow():
    xs_flow_json = TEST_JSON / "xs_flow.json"
    with RasPlanHdf(BAXTER_P01) as phdf:
        assert _gdf_matches_json_alt(phdf.cross_sections_flow(), xs_flow_json)


def test_cross_sections_energy_grade():
    xs_energy_grade_json = TEST_JSON / "xs_energy_grade.json"
    with RasPlanHdf(BAXTER_P01) as phdf:
        assert _gdf_matches_json_alt(
            phdf.cross_sections_energy_grade(), xs_energy_grade_json
        )


def _compare_json(json_file1, json_file2) -> bool:
    with open(json_file1) as j1:
        with open(json_file2) as j2:
            return json.load(j1) == json.load(j2)


def test_zmeta_mesh_cells_timeseries_output(tmp_path):
    with RasPlanHdf(BALD_EAGLE_P18_TIMESERIES) as phdf:
        # Generate Zarr metadata
        zmeta = phdf.zmeta_mesh_cells_timeseries_output("BaldEagleCr")

    # Write the Zarr metadata to JSON
    zmeta_test_path = tmp_path / "bald-eagle-mesh-cells-zmeta.test.json"
    with open(zmeta_test_path, "w") as f:
        json.dump(zmeta, f, indent=4)

    # Compare to a validated JSON file
    zmeta_valid_path = TEST_JSON / "bald-eagle-mesh-cells-zmeta.json"
    assert _compare_json(zmeta_test_path, zmeta_valid_path)

    # Verify that the Zarr metadata can be used to open a dataset
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        backend_kwargs={
            "consolidated": False,
            "storage_options": {"fo": str(zmeta_test_path)},
        },
    )
    assert ds["Water Surface"].shape == (37, 3947)
    assert len(ds.coords["time"]) == 37
    assert len(ds.coords["cell_id"]) == 3947
    assert ds.attrs["mesh_name"] == "BaldEagleCr"


def test_zmeta_mesh_faces_timeseries_output(tmp_path):
    with RasPlanHdf(BALD_EAGLE_P18_TIMESERIES) as phdf:
        # Generate Zarr metadata
        zmeta = phdf.zmeta_mesh_faces_timeseries_output("BaldEagleCr")

    # Write the Zarr metadata to JSON
    zmeta_test_path = tmp_path / "bald-eagle-mesh-faces-zmeta.test.json"
    with open(zmeta_test_path, "w") as f:
        json.dump(zmeta, f, indent=4)

    # Compare to a validated JSON file
    zmeta_valid_path = TEST_JSON / "bald-eagle-mesh-faces-zmeta.json"
    assert _compare_json(zmeta_test_path, zmeta_valid_path)

    # Verify that the Zarr metadata can be used to open a dataset
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        backend_kwargs={
            "consolidated": False,
            "storage_options": {"fo": str(zmeta_test_path)},
        },
    )
    assert ds["Face Velocity"].shape == (37, 7295)
    assert len(ds.coords["time"]) == 37
    assert len(ds.coords["face_id"]) == 7295
    assert ds.attrs["mesh_name"] == "BaldEagleCr"


def test_zmeta_reference_lines_timeseries_output(tmp_path):
    with RasPlanHdf(BALD_EAGLE_P18_REF) as phdf:
        # Generate Zarr metadata
        zmeta = phdf.zmeta_reference_lines_timeseries_output()

    # Write the Zarr metadata to JSON
    zmeta_test_path = tmp_path / "bald-eagle-reflines-zmeta.test.json"
    with open(zmeta_test_path, "w") as f:
        json.dump(zmeta, f, indent=4)

    # Compare to a validated JSON file
    zmeta_valid_path = TEST_JSON / "bald-eagle-reflines-zmeta.json"
    assert _compare_json(zmeta_test_path, zmeta_valid_path)

    # Verify that the Zarr metadata can be used to open a dataset
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        backend_kwargs={
            "consolidated": False,
            "storage_options": {"fo": str(zmeta_test_path)},
        },
    )
    assert ds["Flow"].shape == (37, 4)
    assert len(ds.coords["time"]) == 37
    assert len(ds.coords["refln_id"]) == 4
    assert ds.attrs == {}


def test_zmeta_reference_points_timeseries_output(tmp_path):
    with RasPlanHdf(BALD_EAGLE_P18_REF) as phdf:
        # Generate Zarr metadata
        zmeta = phdf.zmeta_reference_points_timeseries_output()

    # Write the Zarr metadata to JSON
    zmeta_test_path = tmp_path / "bald-eagle-refpoints-zmeta.test.json"
    with open(zmeta_test_path, "w") as f:
        json.dump(zmeta, f, indent=4)

    # Compare to a validated JSON file
    zmeta_valid_path = TEST_JSON / "bald-eagle-refpoints-zmeta.json"
    assert _compare_json(zmeta_test_path, zmeta_valid_path)

    # Verify that the Zarr metadata can be used to open a dataset
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        backend_kwargs={
            "consolidated": False,
            "storage_options": {"fo": str(zmeta_test_path)},
        },
    )
    assert ds["Water Surface"].shape == (37, 3)
    assert ds["Velocity"].shape == (37, 3)
    assert len(ds.coords["time"]) == 37
    assert len(ds.coords["refpt_id"]) == 3
    assert ds.attrs == {}


def test_mesh_cells_summary_output(tmp_path):
    with RasPlanHdf(BALD_EAGLE_P18) as phdf:
        df = phdf.mesh_cells_summary_output()
        test_csv = tmp_path / "BaldEagleDamBrk.summary-cells.test.csv"
        df.to_csv(test_csv)
        filecmp.cmp(
            test_csv,
            TEST_CSV / "BaldEagleDamBrk.summary-cells.csv",
            shallow=False,
        )


def test_mesh_faces_summary_output(tmp_path):
    with RasPlanHdf(BALD_EAGLE_P18) as phdf:
        df = phdf.mesh_faces_summary_output()
        test_csv = tmp_path / "BaldEagleDamBrk.summary-faces.test.csv"
        df.to_csv(test_csv)
        filecmp.cmp(
            test_csv,
            TEST_CSV / "BaldEagleDamBrk.summary-faces.csv",
            shallow=False,
        )


def test__mesh_summary_outputs_df(tmp_path):
    with RasPlanHdf(BALD_EAGLE_P18) as phdf:
        with pytest.raises(ValueError):
            phdf._mesh_summary_outputs_df("neither")

        with pytest.raises(ValueError):
            phdf._mesh_summary_outputs_df(cells_or_faces="cells", output_vars="wrong")

        df = phdf._mesh_summary_outputs_df(
            cells_or_faces="cells",
            output_vars=[
                SummaryOutputVar.MAXIMUM_WATER_SURFACE,
                SummaryOutputVar.MINIMUM_WATER_SURFACE,
            ],
        )
        test_csv = tmp_path / "BaldEagleDamBrk.summary-cells-selectvars.test.csv"
        df.to_csv(test_csv)
        filecmp.cmp(
            test_csv,
            TEST_CSV / "BaldEagleDamBrk.summary-cells-selectvars.csv",
            shallow=False,
        )


def test_observed_timeseries_input_flow():
    with RasPlanHdf(DENTON) as phdf:
        ds = phdf.observed_timeseries_input(vartype="Flow")
        df = ds.sel(refln_name="Denton-Justin_RL").to_dataframe().dropna().reset_index()
        valid_df = pd.read_csv(TEST_CSV / "Denton-Justin_RL_Flow.csv")
        valid_df["time"] = pd.to_datetime(valid_df["time"])
        assert_frame_equal(df, valid_df)


def test_observed_timeseries_input_stage():
    with RasPlanHdf(DENTON) as phdf:
        ds = phdf.observed_timeseries_input(vartype="Stage")
        df = (
            ds.sel(refpt_name="Grapevine_Lake_RP").to_dataframe().dropna().reset_index()
        )
        valid_df = pd.read_csv(TEST_CSV / "Grapevine_Lake_RP_Stage.csv")
        valid_df["time"] = pd.to_datetime(valid_df["time"])
        assert_frame_equal(df, valid_df)


def test_observed_timeseries_input_value_error():
    with RasPlanHdf(DENTON) as phdf:
        with pytest.raises(ValueError):
            phdf.observed_timeseries_input(vartype="Fake Variable")


def test_observed_timeseries_input_rasplanhdf_error():
    with RasPlanHdf(BALD_EAGLE_P18) as phdf:
        with pytest.raises(RasPlanHdfError):
            phdf.observed_timeseries_input(vartype="Flow")
